# Copyright (C) 2008-2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Marking subtitles for easy navigation."""

import aeidon
import gaupol
import gtk
import os
import pango
_ = aeidon.i18n._


class AddBookmarkDialog(gaupol.BuilderDialog):

    """Dialog for editing metadata for a bookmark to be added."""

    _widgets = ("description_entry", "subtitle_spin")

    def __init__(self, parent, page):
        """Initialize an :class:`AddBookmarkDialog` object."""
        directory = os.path.abspath(os.path.dirname(__file__))
        ui_file_path = os.path.join(directory, "add-bookmark-dialog.ui")
        gaupol.BuilderDialog.__init__(self, ui_file_path)
        self._init_values(page)
        self._dialog.set_transient_for(parent)
        self._dialog.set_default_response(gtk.RESPONSE_OK)

    def _init_values(self, page):
        """Initialize default values for widgets."""
        self._subtitle_spin.set_range(1, len(page.project.subtitles))
        row = page.view.get_selected_rows()[0]
        self._subtitle_spin.set_value(row + 1)
        description = page.project.subtitles[row].main_text
        description = description.replace("\n", " ")
        description = aeidon.re_any_tag.sub("", description)
        self._description_entry.set_text(description)
        self._description_entry.set_width_chars(30)

    def get_description(self):
        """Return description of the bookmarked subtitle."""
        return self._description_entry.get_text()

    def get_row(self):
        """Return index of the bookmarked subtitle."""
        return self._subtitle_spin.get_value_as_int() - 1

    def run(self):
        """Show the dialog, run it and return response."""
        self._description_entry.grab_focus()
        return gaupol.BuilderDialog.run(self)


class BookmarksExtension(gaupol.Extension):

    """Marking subtitles for easy navigation."""

    def __init__(self):
        """Initialize a :class:`BookmarksExtension` object."""
        self._action_group = None
        self._bookmarks = None
        self._conf = None
        self._search_entry = None
        self._side_container = None
        self._side_vbox = None
        self._tree_view = None
        self._uim_id = None
        self.application = None

    def _add_bookmark(self):
        """Add a bookmark for the selected subtitle in the current page."""
        page = self.application.get_current_page()
        if not page in self._bookmarks:
            self._bookmarks[page] = {}
        dialog = AddBookmarkDialog(self.application.window, page)
        response = gaupol.util.run_dialog(dialog)
        row = dialog.get_row()
        description = dialog.get_description()
        dialog.destroy()
        if response != gtk.RESPONSE_OK: return
        self._bookmarks[page][row] = description
        self._update_tree_view()
        self.update(self.application, page)

    def _add_bookmark_column(self, page):
        """Add a bookmark column to the subtitle tree view of page."""
        renderer = gtk.CellRendererPixbuf()
        directory = os.path.abspath(os.path.dirname(__file__))
        pixbuf_path = os.path.join(directory, "bookmark.png")
        pixbuf = gtk.gdk.pixbuf_new_from_file(pixbuf_path)
        # Translators: 'Bm.' is short for 'Bookmark'. It is used in the header
        # of a tree view column that contains 16 pixels wide pixbufs.
        column = gtk.TreeViewColumn(_("Bm."), renderer)
        column.set_clickable(True)
        column.set_resizable(False)
        column.set_reorderable(True)
        column.set_visible(self._conf.show_column)
        label = page.view.get_header_label(_("Bm."))
        label.set_tooltip_text(_("Bookmark"))
        column.set_widget(label)
        column.set_data("pixbuf", pixbuf)
        column.set_data("identifier", "bookmark")
        column.set_cell_data_func(renderer, self._set_cell_pixbuf, page)
        page.view.insert_column(column, 0)
        def toggle_bookmark(view, path, column, bookmark_column):
            if column is not bookmark_column: return
            self._toggle_bookmark(int(path[0]))
        page.view.connect("row-activated", toggle_bookmark, column)

    def _clear_attributes(self):
        """Set values of attributes to ``None``."""
        self._action_group = None
        self._bookmarks = None
        self._conf = None
        self._search_entry = None
        self._side_container = None
        self._side_vbox = None
        self._tree_view = None
        self._uim_id = None
        self.application = None

    def _connect_page(self, page):
        """Connect to signals emitted by `page`."""
        page.connect("view-created", self._on_page_view_created)
        callback = self._on_project_subtitles_inserted
        page.project.connect("subtitles-inserted", callback, page)
        callback = self._on_project_subtitles_removed
        page.project.connect("subtitles-removed", callback, page)
        callback = self._on_project_main_file_opened
        page.project.connect("main-file-opened", callback, page)
        callback = self._on_project_main_file_saved
        page.project.connect("main-file-saved", callback, page)

    def _disconnect_application(self, application):
        """Disconnect from signals emitted by ``application``."""
        callback = self._on_application_page_added
        self.application.disconnect("page-added", callback)
        callback = self._on_application_page_closed
        self.application.disconnect("page-closed", callback)
        callback = self._on_application_page_switched
        self.application.disconnect("page-switched", callback)

    def _disconnect_page(self, page):
        """Disconnect from signals emitted by ``page``."""
        callback = self._on_page_view_created
        page.disconnect("view-created", callback)
        callback = self._on_project_subtitles_inserted
        page.project.disconnect("subtitles-inserted", callback)
        callback = self._on_project_subtitles_removed
        page.project.disconnect("subtitles-removed", callback)
        callback = self._on_project_main_file_opened
        page.project.disconnect("main-file-opened", callback)
        callback = self._on_project_main_file_saved
        page.project.disconnect("main-file-saved", callback)

    def _get_bookmark_file_path(self, page):
        """Return the path to ``.gaupol-bookmarks`` file or ``None``."""
        main_file = page.project.main_file
        if main_file is None: return None
        path = main_file.path
        if path.endswith(main_file.format.extension):
            path = path[:-len(main_file.format.extension)]
        return "%s.gaupol-bookmarks" % path

    def _init_actions(self):
        """Initialize UI manager actions."""
        self._action_group.add_actions((
            ("show_bookmarks_menu", None, _("_Bookmarks")),
            ("add_bookmark", gtk.STOCK_ADD, _("_Add\342\200\246"),
             "<Control>D", _("Add a bookmark for the current subtitle"),
             self._on_add_bookmark_activate),
            ("edit_bookmarks", None, _("_Edit Bookmarks"),
             "<Control>B", "Show the bookmarks side pane",
             self._on_edit_bookmarks_activate),
            ("next_bookmark", None, _("_Next"),
             None, _("Go to the next bookmarked subtitle"),
             self._on_next_bookmark_activate),
            ("previous_bookmark", None, _("_Previous"),
             None, _("Go to the previous bookmarked subtitle"),
             self._on_previous_bookmark_activate),))

        self._action_group.add_toggle_actions((
            ("toggle_bookmark_column", None, _("_Bookmark"),
             None, _("Show or hide the bookmark column"),
             self._on_toggle_bookmark_column_toggled,
             self._conf.show_column),))

        self.application.uim.insert_action_group(self._action_group, -1)
        directory = os.path.abspath(os.path.dirname(__file__))
        ui_file_path = os.path.join(directory, "bookmarks.ui.xml")
        self._uim_id = self.application.uim.add_ui_from_file(ui_file_path)
        self.application.uim.ensure_update()
        self.application.set_menu_notify_events("bookmarks")

    def _init_attributes(self, application):
        """Initialize default values for attributes."""
        self._action_group = gtk.ActionGroup("bookmarks")
        self._bookmarks = {}
        self._conf = gaupol.conf.extensions.bookmarks
        self._search_entry = gtk.Entry()
        self._side_container = gtk.Alignment(0, 0, 1, 1)
        self._side_vbox = gtk.VBox(False, 6)
        self._tree_view = gtk.TreeView()
        self._uim_id = None
        self.application = application

    def _init_side_pane_widget(self):
        """Initialize the side pane widget."""
        self._side_vbox.set_border_width(2)
        hbox = gtk.HBox(False, 6)
        label = gtk.Label(_("Search:"))
        hbox.pack_start(label, False, False)
        hbox.pack_start(self._search_entry, True, True)
        self._side_vbox.pack_start(hbox, False, False)
        scroller = gtk.ScrolledWindow()
        scroller.set_policy(*((gtk.POLICY_AUTOMATIC,) * 2))
        scroller.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        scroller.add(self._tree_view)
        self._side_vbox.pack_start(scroller, True, True)
        self._side_container.set_padding(0, 6, 2, 0)
        self._side_container.add(self._side_vbox)
        self._side_container.show_all()

    def _init_signal_handlers(self):
        """Initialize signal handlers."""
        aeidon.util.connect(self, "_search_entry", "changed")
        aeidon.util.connect(self, "_tree_view", "key-press-event")
        aeidon.util.connect(self, "application", "page-added")
        aeidon.util.connect(self, "application", "page-closed")
        aeidon.util.connect(self, "application", "page-switched")

    def _init_tree_view(self):
        """Initialize the side pane tree view."""
        store = gtk.ListStore(bool, int, str)
        store_filter = store.filter_new()
        store_filter.set_visible_column(0)
        self._tree_view.set_model(store_filter)
        self._tree_view.set_headers_visible(False)
        self._tree_view.set_rules_hint(True)
        self._tree_view.set_enable_search(False)
        selection = self._tree_view.get_selection()
        selection.set_mode(gtk.SELECTION_SINGLE)
        selection.connect("changed", self._on_tree_view_selection_changed)
        renderer = gtk.CellRendererText()
        renderer.props.xalign = 1
        column = gtk.TreeViewColumn("", renderer, text=1)
        self._tree_view.append_column(column)
        renderer = gtk.CellRendererText()
        renderer.props.xalign = 0
        renderer.props.editable = True
        renderer.props.ellipsize = pango.ELLIPSIZE_END
        renderer.connect("edited", self._on_tree_view_cell_edited)
        column = gtk.TreeViewColumn("", renderer, text=2)
        self._tree_view.append_column(column)

    def _on_add_bookmark_activate(self, *args):
        """Add a bookmark for the selected subtitle in the current page."""
        self._add_bookmark()

    def _on_application_page_added(self, application, page):
        """Connect to signals in `page` and read bookmarks."""
        self._prepare_page(page)

    def _on_application_page_closed(self, application, page):
        """Remove all data stored for `page`."""
        if page in self._bookmarks:
            del self._bookmarks[page]

    def _on_application_page_switched(self, application, page):
        """Update the side pane to show `page`'s bookmarks."""
        self._update_tree_view()

    def _on_edit_bookmarks_activate(self, *args):
        """Show the bookmarks page in the side pane."""
        action = self.application.get_action("toggle_side_pane")
        action.set_active(True)
        self.application.side_pane.set_current_page(self._side_container)
        self._search_entry.grab_focus()

    def _on_next_bookmark_activate(self, *args):
        """Go to the next bookmarked subtitle."""
        # pylint: disable=W0631
        page = self.application.get_current_page()
        row = page.view.get_selected_rows()[0]
        bookmarks = sorted(self._bookmarks[page].keys())
        bookmarks.append(bookmarks[0])
        for bookmark in bookmarks:
            if bookmark > row: break
        col = page.view.get_focus()[1]
        page.view.set_focus(bookmark, col)
        page.view.scroll_to_row(bookmark)

    def _on_page_view_created(self, page, view):
        """Add a bookmark column to `view`."""
        self._add_bookmark_column(page)

    def _on_project_main_file_opened(self, project, main_file, page):
        """Read bookmarks from ``.gaupol-bookmarks`` file."""
        self._read_bookmarks(page)

    def _on_project_main_file_saved(self, project, main_file, page):
        """Write bookmarks to a ``.gaupol-bookmarks`` file."""
        self._write_bookmarks(page)

    def _on_project_subtitles_inserted(self, project, rows, page):
        """Update rows of bookmarks with `rows` inserted before them."""
        if not page in self._bookmarks: return
        store_filter = self._tree_view.get_model()
        store = store_filter.get_model()
        for irow in rows:
            for crow in sorted(self._bookmarks[page].keys(), reverse=True):
                if crow < irow: continue
                description = self._bookmarks[page][crow]
                del self._bookmarks[page][crow]
                self._bookmarks[page][crow + 1] = description
        self._update_tree_view()

    def _on_project_subtitles_removed(self, project, rows, page):
        """Remove bookmarks and update rows of those remaining."""
        if not page in self._bookmarks: return
        store_filter = self._tree_view.get_model()
        store = store_filter.get_model()
        for crow in self._bookmarks[page].keys():
            if crow in rows: del self._bookmarks[page][crow]
        for crow in sorted(self._bookmarks[page].keys()):
            count = sum(1 for x in rows if x < crow)
            if count == 0: continue
            description = self._bookmarks[page][crow]
            del self._bookmarks[page][crow]
            self._bookmarks[page][crow - count] = description
        self._update_tree_view()

    def _on_previous_bookmark_activate(self, *args):
        """Go to the previous bookmarked subtitle."""
        # pylint: disable=W0631
        page = self.application.get_current_page()
        row = page.view.get_selected_rows()[0]
        bookmarks = sorted(self._bookmarks[page].keys())
        bookmarks.insert(0, bookmarks[-1])
        for bookmark in reversed(bookmarks):
            if bookmark < row: break
        col = page.view.get_focus()[1]
        page.view.set_focus(bookmark, col)
        page.view.scroll_to_row(bookmark)

    def _on_search_entry_changed(self, entry):
        """Check which bookmark descriptions match with new search string."""
        store_filter = self._tree_view.get_model()
        store = store_filter.get_model()
        pattern = entry.get_text().lower()
        for i, (visible, number, description) in enumerate(store):
            store[i][0] = (description.lower().find(pattern) >= 0)

    def _on_toggle_bookmark_column_toggled(self, action, *args):
        """Show or hide the bookmark column."""
        page = self.application.get_current_page()
        col = page.view.columns.BOOKMARK
        page.view.get_column(col).set_visible(action.get_active())
        self._conf.show_column = action.get_active()

    def _on_tree_view_cell_edited(self, renderer, path, new_text):
        """Update description in the list store model of the tree view."""
        store = self._tree_view.get_model().get_model()
        store[path][2] = unicode(new_text)
        page = self.application.get_current_page()

    def _on_tree_view_key_press_event(self, tree_view, event):
        """Remove selected bookmark if ``Delete`` key pressed."""
        if event.keyval != gtk.keysyms.Delete: return
        selection = self._tree_view.get_selection()
        store, tree_iter = selection.get_selected()
        if tree_iter is None: return
        path = store.get_path(tree_iter)
        self._remove_bookmark(store[path][1] - 1)

    def _on_tree_view_selection_changed(self, selection):
        """Jump to the subtitle of the selected bookmark."""
        store, tree_iter = selection.get_selected()
        if tree_iter is None: return
        path = store.get_path(tree_iter)
        row = store[path][1] - 1
        page = self.application.get_current_page()
        col = page.view.get_focus()[1]
        page.view.set_focus(row, col)
        page.view.scroll_to_row(row)

    def _prepare_page(self, page):
        """Prepare data structures and signal handlers for `page`."""
        if not page in self._bookmarks:
            self._bookmarks[page] = {}
        self._connect_page(page)
        self._add_bookmark_column(page)
        self._read_bookmarks(page)

    def _read_bookmarks(self, page):
        """Read bookmarks for `page` from ``.gaupol-bookmarks`` file."""
        if not page in self._bookmarks:
            self._bookmarks[page] = {}
        self._bookmarks[page].clear()
        path = self._get_bookmark_file_path(page)
        if path is None: return
        if not os.path.isfile(path): return
        encoding = page.project.main_file.encoding
        for line in aeidon.util.readlines(path, encoding):
            row, description = line.split(" ", 1)
            self._bookmarks[page][int(row) - 1] = description
        self._update_tree_view()

    def _remove_bookmark(self, row):
        """Remove existing bookmark in `row` of the current page."""
        store = self._tree_view.get_model()
        path = None
        for i in range(len(store)):
            if store[i][1] - 1 == row:
                path = i
        if path is not None:
            del store[path]
        page = self.application.get_current_page()
        if row in self._bookmarks[page]:
            del self._bookmarks[page][row]
        # Update the pixbuf column immediately.
        page_store = page.view.get_model()
        page_store.row_changed(row, page_store.get_iter(row))
        self.update(self.application, page)

    def _set_cell_pixbuf(self, column, renderer, store, tree_iter, page):
        """Set the pixbuf property of `renderer`."""
        pixbuf = None
        if page in self._bookmarks:
            row = store.get_path(tree_iter)[0]
            if row in self._bookmarks[page]:
                pixbuf = column.get_data("pixbuf")
        renderer.props.pixbuf = pixbuf

    def _toggle_bookmark(self, row):
        """Remove existing bookmark or add new bookmark to the current page."""
        page = self.application.get_current_page()
        if row in self._bookmarks[page]:
            return self._remove_bookmark(row)
        return self._add_bookmark()

    def _update_tree_view(self):
        """Update tree view to display bookmarks for the current page."""
        store_filter = self._tree_view.get_model()
        store = store_filter.get_model()
        store.clear()
        page = self.application.get_current_page()
        if not page in self._bookmarks: return
        pattern = self._search_entry.get_text().lower()
        for row in sorted(self._bookmarks[page].keys()):
            description = self._bookmarks[page][row]
            visible = (description.lower().find(pattern) >= 0)
            store.append((visible, row + 1, description))

    def _write_bookmarks(self, page):
        """Write bookmarks from `page` to a ``.gaupol-bookmarks`` file."""
        if not page in self._bookmarks: return
        path = self._get_bookmark_file_path(page)
        if path is None: return
        if not self._bookmarks[page]:
            # Remove bookmark file when all bookmarks are removed.
            if os.path.isfile(path): os.remove(path)
            return
        lines = []
        for row in sorted(self._bookmarks[page].keys()):
            lines.append("%d %s" % (row + 1, self._bookmarks[page][row]))
        text = os.linesep.join(lines) + os.linesep
        encoding = page.project.main_file.encoding
        aeidon.util.write(path, text, encoding)

    def setup(self, application):
        """Setup extension for use with `application`."""
        gaupol.conf.register_extension("bookmarks", {"show_column": True})
        self._init_attributes(application)
        self._init_tree_view()
        self._init_signal_handlers()
        self._init_side_pane_widget()
        self._init_actions()
        for page in application.pages:
            self._prepare_page(page)
        application.side_pane.add_page(self._side_container,
                                       "bookmarks",
                                       _("Bookmarks"))

    def teardown(self, application):
        """End use of extension with `application`."""
        application.uim.remove_ui(self._uim_id)
        application.uim.remove_action_group(self._action_group)
        application.uim.ensure_update()
        for page in application.pages:
            col = page.view.columns.BOOKMARK
            column = page.view.get_column(col)
            page.view.remove_column(column)
            self._disconnect_page(page)
        self._disconnect_application(application)
        application.side_pane.remove_page(self._side_container)
        self._clear_attributes()

    def update(self, application, page):
        """Update state of extension for `application` and active `page`."""
        action = self._action_group.get_action("add_bookmark")
        try: action.set_sensitive(len(page.view.get_selected_rows()) == 1)
        except AttributeError: action.set_sensitive(False)
        action = self._action_group.get_action("edit_bookmarks")
        action.set_sensitive(page is not None)
        action = self._action_group.get_action("next_bookmark")
        try: action.set_sensitive(bool(self._bookmarks[page]))
        except KeyError: action.set_sensitive(False)
        action = self._action_group.get_action("previous_bookmark")
        try: action.set_sensitive(bool(self._bookmarks[page]))
        except KeyError: action.set_sensitive(False)



# Copyright (C) 2009-2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Cell renderer for float data with fixed precision."""

import aeidon
import glib
import gtk

__all__ = ("FloatCellRenderer",)


class FloatCellRenderer(gtk.CellRendererText):

    """Cell renderer for float data with fixed precision."""

    __gtype_name__ = "FloatCellRenderer"

    def __init__(self, format="%.3f"):
        """Initialize a :class:`FloatCellRenderer` object."""
        gtk.CellRendererText.__init__(self)
        self._format = format
        self._text = ""
        aeidon.util.connect(self, self, "notify::text")

    def _on_notify_text(self, *args):
        """Cut decimals to fixed precision."""
        self._text = text = unicode(self.props.text)
        if not text: return
        has_comma = text.find(",") > 0
        text = text.replace(",", ".")
        text = self._format % float(text)
        text = (text.replace(".", ",") if has_comma else text)
        self.props.markup = glib.markup_escape_text(text)



# Copyright (C) 2005-2008,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import aeidon
import gaupol
import gtk


class TestHelpAgent(gaupol.TestCase):

    def setup_method(self, method):
        self.application = self.new_application()

    def test_on_browse_wiki_documentation_activate(self):
        self.application.get_action("browse_wiki_documentation").activate()

    def test_on_report_a_bug_activate(self):
        self.application.get_action("report_a_bug").activate()

    @aeidon.deco.monkey_patch(gaupol.util, "flash_dialog")
    def test_on_view_about_dialog_activate(self):
        respond = lambda *args: gtk.RESPONSE_DELETE_EVENT
        gaupol.util.flash_dialog = respond
        self.application.get_action("view_about_dialog").activate()



# Copyright (C) 2007-2008,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

""":class:`gtk.UIManager` actions for :class:`gaupol.Application`."""

import aeidon

from gaupol.actions.edit import *
from gaupol.actions.file import *
from gaupol.actions.format import *
from gaupol.actions.help import *
from gaupol.actions.menu import *
from gaupol.actions.position import *
from gaupol.actions.search import *
from gaupol.actions.text import *
from gaupol.actions.view import *

__all__ = tuple(filter(lambda x: x.endswith("Action"), dir()))



# Copyright (C) 2005-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Lists of named constants with integer values.

:class:`EnumerationItem`s are subclassed from :class:`int`, so they are actual
integers. The string value of an item will be the name that it was defined with
in its set. New items can always be added to an enumeration.
"""

import aeidon

__all__ = ("EnumerationItem", "Enumeration",)


class EnumerationItem(int):

    """Named constant with an integer value.

    This class can be instantiated without any specified values and users need
    not bother with them. Instantiation with correct values harmonous with the
    rest of the items should be left up to the parent list.
    """

    # pylint: disable=E1101

    def __bool__(self):
        """For consistency, always return True."""
        return True

    def __cmp__(self, other):
        """Compare enumeration item equality by value.

        If `other` is an integer, return integer comparison value. Raise
        :exc:`ValueError` if `other` is an :class:`EnumerationItem` of a
        different parent. If `other` is not an integer, return -1. This is a
        debug method that exists only if :data:`aeidon.debug` is ``True``.
        """
        if isinstance(other, int):
            if isinstance(other, EnumerationItem):
                if self.parent is not other.parent:
                    raise ValueError("Cannot compare %s with %s"
                                     % (repr(self), repr(other)))

            return int.__cmp__(int(self), int(other))
        if other is None:
            return -1
        raise ValueError("Cannot compare %s with %s"
                         % (repr(self), repr(other)))

    if not aeidon.debug: del __cmp__

    def __new__(cls, value=0, name="", parent=None):
        """Return integer instance with additional attributes."""
        instance = int.__new__(cls, value)
        instance.name = name
        instance.parent = parent
        return instance

    def __nonzero__(self):
        """For consistency, always return True."""
        return True

    def __str__(self):
        """Return name as the string representation."""
        return self.name


class Enumeration(list):

    """List of named constants with integer values.

    This class is an actual :class:`list` where enumeration items are stored as
    both list items and instance attributes. New items should be added by
    setting an instance attribute. Data-changing list methods raise
    :exc:`NotImplementedError`.

    Typical use to create a new enumeration would be something like::

        fruits = aeidon.Enumeration()
        fruits.APPLE = aeidon.EnumerationItem()
        fruits.APPLE.size = 10
        fruits.MANGO = aeidon.EnumerationItem()
        fruits.MANGO.size = 15

    Note that there is no finalization of an enumeration. New items can always
    be added just by assigning a new attribute to the enumeration. Likewise,
    existing items can always be removed using :func:`delattr`.
    """

    NONE = None

    def __contains__(self, item):
        if isinstance(item, EnumerationItem):
            if item.parent is not self:
                return False
        return list.__contains__(self, item)

    if not aeidon.debug: del __contains__

    def __delattr__(self, name):
        """Delete enumeration item and attribute."""
        list.remove(self, getattr(self, name))
        return object.__delattr__(self, name)

    def __delitem__(self, *args, **kwargs):
        raise NotImplementedError

    def __delslice__(self, *args, **kwargs):
        raise NotImplementedError

    def __iadd__(self, *args, **kwargs):
        raise NotImplementedError

    def __imul__(self, *args, **kwargs):
        raise NotImplementedError

    def __setattr__(self, name, value):
        """Set value of enumeration item with correct attributes."""
        if isinstance(value, EnumerationItem):
            value = value.__class__(len(self), name, self)
            list.append(self, value)
        return object.__setattr__(self, name, value)

    def __setitem__(self, *args, **kwargs):
        raise NotImplementedError

    def __setslice__(self, *args, **kwargs):
        raise NotImplementedError

    def append(self, *args, **kwargs):
        raise NotImplementedError

    def extend(self, *args, **kwargs):
        raise NotImplementedError

    def find_item(self, name, value):
        """Return the first found item with the given attribute `value`."""
        for item in self:
            if getattr(item, name) == value:
                return item
        raise ValueError("Name %s not found" % repr(name))

    def insert(self, *args, **kwargs):
        raise NotImplementedError

    def pop(self, *args, **kwargs):
        raise NotImplementedError

    def remove(self, *args, **kwargs):
        raise NotImplementedError

    def reverse(self, *args, **kwargs):
        raise NotImplementedError

    def sort(self, *args, **kwargs):
        raise NotImplementedError



# Copyright (C) 2005-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Names and codes for locales and conversions between them.

Locale codes are of form ``aa[_BB][@Cccc]``, where ``aa`` is a language code,
``BB`` a country code and ``Cccc`` a script. See :mod:`aeidon.languages`,
:mod:`aeidon.countries` and :mod:`aeidon.scripts` for the details.
"""

import aeidon
import os
import re
_ = aeidon.i18n._


def code_to_country_require(code):
    assert re.match(r"^[a-z][a-z](_[A-Z][A-Z])?$", code)

@aeidon.deco.contractual
def code_to_country(code):
    """Convert locale `code` to localized country name or ``None``.

    Raise :exc:`KeyError` if `code` not found.
    """
    if len(code) < 5: return None
    return aeidon.countries.code_to_name(code[-2:])

def code_to_language_require(code):
    assert re.match(r"^[a-z][a-z](_[A-Z][A-Z])?$", code)

@aeidon.deco.contractual
def code_to_language(code):
    """Convert locale `code` to localized language name.

    Raise :exc:`KeyError` if `code` not found.
    """
    return aeidon.languages.code_to_name(code[:2])

def code_to_name_require(code):
    assert re.match(r"^[a-z][a-z](_[A-Z][A-Z])?$", code)

@aeidon.deco.contractual
def code_to_name(code):
    """Convert locale `code` to localized name.

    Raise :exc:`KeyError` if `code` not found.
    Return localized ``LANGUAGE (COUNTRY)``.
    """
    language = code_to_language(code)
    country = code_to_country(code)
    if country is None: return language
    return _("%(language)s (%(country)s)") % locals()

@aeidon.deco.once
def get_system_code():
    """Return the locale code preferred by system or ``None``."""
    import locale
    return locale.getdefaultlocale()[0]

@aeidon.deco.once
def get_system_modifier():
    """Return the script modifier of system or ``None``."""
    import locale
    names = ("LANGUAGE", "LC_ALL", "LC_MESSAGES", "LANG")
    values = map(os.environ.get, names)
    values = filter(None, values)
    for value in (x for x in values if x.count("@")):
        return value[value.index("@") + 1:value.index("@") + 5]
    return None



# Copyright (C) 2005-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import aeidon


class TestClipboardAgent(aeidon.TestCase):

    def setup_method(self, method):
        self.project = self.new_project()

    def test_copy_texts(self):
        text_0 = self.project.subtitles[0].main_text
        text_2 = self.project.subtitles[2].main_text
        self.project.copy_texts((0, 2), aeidon.documents.MAIN)
        texts = self.project.clipboard.get_texts()
        assert texts == [text_0, None, text_2]
        assert self.project.subtitles[0].main_text == text_0
        assert self.project.subtitles[2].main_text == text_2

    @aeidon.deco.reversion_test
    def test_cut_texts(self):
        text_0 = self.project.subtitles[0].main_text
        text_2 = self.project.subtitles[2].main_text
        self.project.cut_texts((0, 2), aeidon.documents.MAIN)
        texts = self.project.clipboard.get_texts()
        assert texts == [text_0, None, text_2]
        assert self.project.subtitles[0].main_text == ""
        assert self.project.subtitles[2].main_text == ""

    @aeidon.deco.reversion_test
    def test_paste_texts__fit(self):
        subtitles = self.project.subtitles
        self.project.copy_texts((0, 1), aeidon.documents.TRAN)
        indices = self.project.paste_texts(2, aeidon.documents.MAIN)
        assert indices == (2, 3)
        assert subtitles[0].tran_text == subtitles[2].main_text
        assert subtitles[1].tran_text == subtitles[3].main_text

    @aeidon.deco.reversion_test
    def test_paste_texts__new(self):
        subtitles = self.project.subtitles
        self.project.copy_texts((0, 1), aeidon.documents.TRAN)
        z = len(subtitles) - 1
        indices = self.project.paste_texts(z, aeidon.documents.MAIN)
        assert list(indices) == [z, z + 1]
        assert len(subtitles) == z + 2



# Copyright (C) 2005-2008,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import aeidon
import gaupol
import gtk


class TestEditAgent(gaupol.TestCase):

    def setup_method(self, method):
        self.application = self.new_application()
        self.delegate = self.application.undo.im_self

    def test_on_clear_texts_activate(self):
        page = self.application.get_current_page()
        page.view.set_focus(0, page.view.columns.MAIN_TEXT)
        page.view.select_rows((0, 1, 2))
        self.application.get_action("clear_texts").activate()

    def test_on_copy_texts_activate(self):
        page = self.application.get_current_page()
        page.view.set_focus(0, page.view.columns.MAIN_TEXT)
        page.view.select_rows((0, 1, 2))
        self.application.get_action("copy_texts").activate()

    def test_on_cut_texts_activate(self):
        page = self.application.get_current_page()
        page.view.set_focus(0, page.view.columns.MAIN_TEXT)
        page.view.select_rows((0, 1, 2))
        self.application.get_action("cut_texts").activate()

    @aeidon.deco.monkey_patch(gaupol.util, "flash_dialog")
    def test_on_edit_headers_activate(self):
        gaupol.util.flash_dialog = lambda *args: gtk.RESPONSE_OK
        path = self.new_temp_file(aeidon.formats.SUBVIEWER2)
        self.application.open_main(path)
        self.application.get_action("edit_headers").activate()

    def test__on_edit_next_value_activate(self):
        page = self.application.get_current_page()
        page.view.set_focus(0, page.view.columns.MAIN_TEXT)
        self.application.get_action("edit_next_value").activate()

    def test_on_edit_preferences_activate(self):
        self.application.get_action("edit_preferences").activate()
        self.application.get_action("edit_preferences").activate()
        self.delegate._pref_dialog.response(gtk.RESPONSE_CLOSE)

    def test__on_edit_value_activate(self):
        page = self.application.get_current_page()
        page.view.set_focus(0, page.view.columns.MAIN_TEXT)
        self.application.get_action("edit_value").activate()

    def test_on_extend_selection_to_beginning_activate(self):
        page = self.application.get_current_page()
        page.view.select_rows((4, 5))
        self.application.get_action("extend_selection_to_beginning").activate()

    def test_on_extend_selection_to_end_activate(self):
        page = self.application.get_current_page()
        page.view.select_rows((4, 5))
        self.application.get_action("extend_selection_to_end").activate()

    @aeidon.deco.monkey_patch(gaupol.util, "flash_dialog")
    def test_on_insert_subtitles_activate(self):
        gaupol.util.flash_dialog = lambda *args: gtk.RESPONSE_OK
        page = self.application.get_current_page()
        page.view.set_focus(0, page.view.columns.MAIN_TEXT)
        self.application.get_action("insert_subtitles").activate()

    def test_on_invert_selection_activate(self):
        self.application.get_action("invert_selection").activate()

    def test_on_merge_subtitles_activate(self):
        page = self.application.get_current_page()
        page.view.set_focus(0, page.view.columns.MAIN_TEXT)
        page.view.select_rows((0, 1))
        self.application.get_action("merge_subtitles").activate()

    def test_on_paste_texts_activate(self):
        page = self.application.get_current_page()
        page.view.set_focus(0, page.view.columns.MAIN_TEXT)
        page.view.select_rows((0, 1, 2))
        self.application.get_action("copy_texts").activate()
        page.view.set_focus(0, page.view.columns.MAIN_TEXT)
        self.application.get_action("paste_texts").activate()

    def test_on_paste_texts_activate__subtitles_added(self):
        page = self.application.get_current_page()
        page.view.set_focus(0, page.view.columns.MAIN_TEXT)
        page.view.select_rows((0, 1, 2))
        self.application.get_action("copy_texts").activate()
        row = len(page.project.subtitles) - 1
        page.view.set_focus(row, page.view.columns.MAIN_TEXT)
        self.application.get_action("paste_texts").activate()

    def test_on_project_action_done(self):
        page = self.application.get_current_page()
        page.project.remove_subtitles((0,))

    def test_on_project_action_redone(self):
        page = self.application.get_current_page()
        page.project.remove_subtitles((0,))
        self.application.undo()
        self.application.redo()

    def test_on_project_action_undone(self):
        page = self.application.get_current_page()
        page.project.remove_subtitles((0,))
        self.application.undo()

    def test_on_redo_action_activate(self):
        page = self.application.get_current_page()
        page.project.remove_subtitles((0,))
        self.application.get_action("undo_action").activate()
        self.application.get_action("redo_action").activate()

    def test_on_remove_subtitles_activate(self):
        page = self.application.get_current_page()
        page.view.select_rows((0, 1, 2))
        self.application.get_action("remove_subtitles").activate()

    def test_on_select_all_activate(self):
        self.application.get_action("select_all").activate()

    def test_on_split_subtitle_activate(self):
        page = self.application.get_current_page()
        page.view.set_focus(0, page.view.columns.MAIN_TEXT)
        self.application.get_action("split_subtitle").activate()

    def test_on_undo_action_activate(self):
        page = self.application.get_current_page()
        page.project.remove_subtitles((0,))
        self.application.get_action("undo_action").activate()

    def test_on_view_renderer_edited__position_frame(self):
        page = self.application.get_current_page()
        self.application.get_action("show_frames").activate()
        for col in filter(page.view.is_position_column, page.view.columns):
            column = page.view.get_column(col)
            renderer = column.get_cell_renderers()[0]
            renderer.emit("edited", 1, 0)
            renderer.emit("edited", 1, "k")

    def test_on_view_renderer_edited__position_time(self):
        page = self.application.get_current_page()
        self.application.get_action("show_times").activate()
        for col in filter(page.view.is_position_column, page.view.columns):
            column = page.view.get_column(col)
            renderer = column.get_cell_renderers()[0]
            renderer.emit("edited", 1, "00:00:00.000")

    def test_on_view_renderer_edited__text(self):
        page = self.application.get_current_page()
        self.application.get_action("show_times").activate()
        for col in filter(page.view.is_text_column, page.view.columns):
            column = page.view.get_column(col)
            renderer = column.get_cell_renderers()[0]
            renderer.emit("edited", 1, "test")

    def test_on_view_renderer_editing_canceled__position(self):
        page = self.application.get_current_page()
        for col in filter(page.view.is_position_column, page.view.columns):
            column = page.view.get_column(col)
            renderer = column.get_cell_renderers()[0]
            renderer.emit("editing-canceled")

    def test_on_view_renderer_editing_canceled__text(self):
        page = self.application.get_current_page()
        for col in filter(page.view.is_text_column, page.view.columns):
            column = page.view.get_column(col)
            renderer = column.get_cell_renderers()[0]
            renderer.emit("editing-canceled")

    def test_on_view_renderer_editing_started__position(self):
        page = self.application.get_current_page()
        for col in filter(page.view.is_text_column, page.view.columns):
            column = page.view.get_column(col)
            page.view.set_cursor(1, column, True)
            gaupol.util.iterate_main()

    def test_on_view_renderer_editing_started__position(self):
        page = self.application.get_current_page()
        for col in filter(page.view.is_position_column, page.view.columns):
            column = page.view.get_column(col)
            page.view.set_cursor(1, column, True)
            gaupol.util.iterate_main()

    def test_redo(self):
        page = self.application.get_current_page()
        page.project.remove_subtitles((0,))
        self.application.undo()
        self.application.redo()

    def test_undo(self):
        page = self.application.get_current_page()
        page.project.remove_subtitles((0,))
        self.application.undo()



# Copyright (C) 2005-2007,2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""All :mod:`aeidon` error classes."""

__all__ = ("Error",
           "AffirmationError",
           "FormatError",
           "ParseError",
           "ProcessError",)


class Error(Exception):

    """Base class for all :mod:`aeidon` errors."""

    pass


class AffirmationError(Error):

    """Something expected to be ``True`` was ``False``.

    :exc:`AffirmationError` is by nature similar to the built-in
    :exc:`AssertionError`, but without the special reliance on
    :const:`__debug__` and given optimization options. :exc:`AffirmationError`
    is used to provide essential checks of boolean values instead of optional
    debug checks.
    """

    pass


class FormatError(Error):

    """Unrecognized subtitle file format."""

    pass


class ParseError(Error):

    """Failed to parse a subtitle file into positions and texts.

    :exc:`ParseError` is caused by either a syntax error in the file being
    parsed or a programming error in the code doing the parsing.
    """

    pass


class ProcessError(Error):

    """Error in an external process."""

    pass



# Copyright (C) 2005-2008,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Dialog for selecting a video file."""

import aeidon
import gtk
_ = aeidon.i18n._

__all__ = ("VideoDialog",)


class VideoDialog(gtk.FileChooserDialog):

    """Dialog for selecting a video file."""

    def __init__(self, parent):
        """Initialize a :class:`VideoDialog` object."""
        gtk.FileChooserDialog.__init__(self)
        self.set_title(_("Select Video"))
        self.set_transient_for(parent)
        self.set_action(gtk.FILE_CHOOSER_ACTION_OPEN)
        self.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
        self.add_button(_("_Select"), gtk.RESPONSE_OK)
        self.set_default_response(gtk.RESPONSE_OK)
        self._init_filters()

    def _init_filters(self):
        """Intialize the file filters."""
        file_filter = gtk.FileFilter()
        file_filter.add_pattern("*")
        file_filter.set_name(_("All files"))
        self.add_filter(file_filter)
        file_filter = gtk.FileFilter()
        file_filter.add_mime_type("video/*")
        file_filter.set_name(_("Video files"))
        self.add_filter(file_filter)
        self.set_filter(file_filter)



# Copyright (C) 2005-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Writing subtitle data to file."""

import aeidon
import os
import shutil
import sys


class SaveAgent(aeidon.Delegate):

    """Writing subtitle data to file."""

    # pylint: disable=E0203,E1101,W0201

    __metaclass__ = aeidon.Contractual

    def _convert_markup(self, doc, from_format, to_format):
        """Convert markup in texts and return changed indices."""
        indices = []
        converter = aeidon.MarkupConverter(from_format, to_format)
        for i, subtitle in enumerate(self.subtitles):
            text = subtitle.get_text(doc)
            new_text = converter.convert(text)
            if new_text == text: continue
            subtitle.set_text(doc, new_text)
            indices.append(i)
        return indices

    def _copy_file_ensure(self, value, source, destination):
        assert (not value) or os.path.isfile(destination)

    def _copy_file(self, source, destination):
        """Copy source file to destination and return success."""
        try:
            shutil.copyfile(source, destination)
            return True
        except IOError:
            aeidon.util.print_write_io(sys.exc_info(), destination)
        return False

    def _ensure_mode(self, mode):
        """Update mode of subtitles if necessary."""
        if self.main_file is None: return
        if mode == self.main_file.mode: return
        for i, subtitle in enumerate(self.subtitles):
            subtitle.mode = mode
        self.emit("positions-changed", self.get_all_indices())

    def _move_file_ensure(self, value, source, destination):
        assert (not value) or os.path.isfile(destination)

    def _move_file(self, source, destination):
        """Move source file to destination and return success."""
        try:
            shutil.move(source, destination)
            return True
        except (IOError, OSError):
            aeidon.util.print_write_io(sys.exc_info(), destination)
        return False

    def _remove_file_ensure(self, value, path):
        assert (not value) or (not os.path.isfile(path))

    def _remove_file(self, path):
        """Remove file and return success."""
        try:
            os.remove(path)
            return True
        except OSError:
            aeidon.util.print_remove_os(sys.exc_info(), path)
        return False

    def _save(self, doc, sfile, keep_changes):
        """Write subtitle data from `doc` to `sfile`.

        Raise :exc:`IOError` if writing fails.
        Raise :exc:`UnicodeError` if encoding fails.
        Return indices of texts changed due to markup conversion.
        """
        current_file = self.get_file(doc)
        current_format = self.get_format(doc)
        orig_texts = [x.get_text(doc) for x in self.subtitles]
        indices = []
        if (current_format is not None) and (sfile.format != current_format):
            indices = self._convert_markup(doc, current_format, sfile.format)
        self._write_file(sfile, self.subtitles, doc)
        if keep_changes: return indices
        for i, subtitle in enumerate(self.subtitles):
            subtitle.set_text(doc, orig_texts[i])
        return []

    def _write_file_ensure(self, value, sfile, subtitles, doc):
        assert os.path.isfile(sfile.path)

    def _write_file(self, sfile, subtitles, doc):
        """Write subtitle data to `sfile`.

        Raise :exc.`IOError` if writing fails.
        Raise :exc.`UnicodeError` if encoding fails.
        """
        file_existed = os.path.isfile(sfile.path)
        if os.path.isfile(sfile.path):
            # Create a backup of existing file in case writing
            # new data to that file fails.
            backup_path = aeidon.temp.create(".bak")
            aeidon.temp.close(backup_path)
            backup_success = self._copy_file(sfile.path, backup_path)
        try: sfile.write(subtitles, doc)
        except (IOError, UnicodeError):
            if file_existed and backup_success:
                # If overwriting existing file fails,
                # restore old file from backup.
                self._move_file(backup_path, sfile.path)
            if not file_existed:
                # If writing new file fails, remove failed attempt,
                # which is a blank or an incomplete file.
                self._remove_file(sfile.path)
            raise # (IOError, UnicodeError)
        if file_existed and backup_success:
            aeidon.temp.remove(backup_path)

    @aeidon.deco.export
    def save(self, doc, sfile=None, keep_changes=True):
        """Write subtitle data from `doc` to `sfile`.

        `sfile` can be ``None`` to use existing file, i.e. :attr:`main_file`
        or :attr:`tran_file`.
        Raise :exc:`IOError` if writing fails.
        Raise :exc:`UnicodeError` if encoding fails.
        """
        if doc == aeidon.documents.MAIN:
            return self.save_main(sfile, keep_changes)
        if doc == aeidon.documents.TRAN:
            return self.save_translation(sfile, keep_changes)
        raise ValueError("Invalid document: %s" % repr(doc))

    def save_main_ensure(self, value, sfile=None, keep_changes=True):
        assert self.main_file is not None
        assert (not keep_changes) or (self.main_changed == 0)

    @aeidon.deco.export
    def save_main(self, sfile=None, keep_changes=True):
        """Write subtitle data from main document to `sfile`.

        `sfile` can be ``None`` to use :attr:`main_file`.
        Raise :exc:`IOError` if writing fails.
        Raise :exc:`UnicodeError` if encoding fails.
        """
        sfile = sfile or self.main_file
        if sfile is not None and self.main_file is not None:
            sfile.copy_from(self.main_file)
        indices = self._save(aeidon.documents.MAIN, sfile, keep_changes)
        if keep_changes:
            self._ensure_mode(sfile.mode)
            self.main_file = sfile
            self.main_changed = 0
            self.emit("main-texts-changed", indices)
        self.emit("main-file-saved", sfile)

    def save_translation_ensure(self, value, sfile=None, keep_changes=True):
        assert self.tran_file is not None
        assert (not keep_changes) or (self.tran_changed == 0)

    @aeidon.deco.export
    def save_translation(self, sfile=None, keep_changes=True):
        """Write subtitle data from translation document to `sfile`.

        `sfile` can be ``None`` to use :attr:`tran_file`.
        Raise :exc:`IOError` if writing fails.
        Raise :exc:`UnicodeError` if encoding fails.
        """
        sfile = sfile or self.tran_file
        if sfile is not None and self.tran_file is not None:
            sfile.copy_from(self.tran_file)
        indices = self._save(aeidon.documents.TRAN, sfile, keep_changes)
        if keep_changes:
            self.tran_file = sfile
            self.tran_changed = 0
            self.emit("translation-texts-changed", indices)
        self.emit("translation-file-saved", sfile)



# Copyright (C) 2007-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Managing regular expression substitutions for subtitle texts."""

import aeidon
import os
import re
import xml.etree.ElementTree as ET

__all__ = ("PatternManager",)


class PatternManager(object):

    """Managing regular expression substitutions for subtitle texts.

    :ivar _patterns: Dictionary mapping codes to pattern lists
    :ivar pattern_type: String to indentify what the pattern matches

    :attr:`pattern_type` should be a string with value "line-break",
    "common-error", "capitalization" or "hearing-impaired". Codes are of form
    ``Script[-language-[COUNTRY]]`` using the corresponding ISO codes.
    """

    __metaclass__ = aeidon.Contractual
    _re_comment = re.compile(r"^\s*#.*$")

    def __init___require(self, pattern_type):
        assert pattern_type in ("capitalization",
                                "common-error",
                                "hearing-impaired",
                                "line-break")

    def __init__(self, pattern_type):
        """Initialize a :class:`PatternManager` object."""
        self.pattern_type = pattern_type
        self._patterns = {}
        self._read_patterns()

    def _assert_identifiers(self, script, language, country):
        """Assert that codes are valid and parents are defined."""
        if script is not None:
            assert aeidon.scripts.is_valid(script)
        if language is not None:
            assert script is not None
            assert aeidon.languages.is_valid(language)
        if country is not None:
            assert language is not None
            assert aeidon.countries.is_valid(country)

    def _filter_patterns(self, patterns):
        """Return `patterns` with name-clashes resolved.

        Patterns with a more specific code replace those with a less specific
        code if they have the same name and the more specific pattern's policy
        is explicitly set to ``Replace`` (instead of the implicit ``Append``).
        Order is mainteined so that all patterns with the same name are always
        located in the position of the earliest of such patterns.
        """
        filtered_patterns = []
        for i, pattern in enumerate(patterns):
            name = pattern.get_name(False)
            policy = pattern.get_field("Policy")
            last_index = len(filtered_patterns) - 1
            for j, filtered_pattern in enumerate(filtered_patterns):
                if filtered_pattern.get_name() == name:
                    last_index = j
                    if policy == "Replace":
                        filtered_patterns[j] = None
            filtered_patterns.insert(last_index + 1, pattern)
            while None in filtered_patterns:
                filtered_patterns.remove(None)
        return filtered_patterns

    def _get_codes_require(self, script=None, language=None, country=None):
        self._assert_identifiers(script, language, country)

    def _get_codes(self, script=None, language=None, country=None):
        """Return a sequence of all codes to be used by arguments.

        ``Zyyy`` is the first code and the most specific one last.
        """
        codes = ["Zyyy"]
        if script is not None:
            codes.append(script)
        if language is not None:
            code = "%s-%s" % (script, language)
            codes.append(code)
        if country is not None:
            code = "%s-%s" % (code, country)
            codes.append(code)
        return tuple(codes)

    def _read_config_from_directory_require(self, directory, encoding):
        assert aeidon.encodings.is_valid_code(encoding)

    def _read_config_from_directory(self, directory, encoding):
        """Read configurations from files in `directory`."""
        if not os.path.isdir(directory): return
        extension = ".%s.conf" % self.pattern_type
        files = os.listdir(directory)
        for name in (x for x in files if x.endswith(extension)):
            path = os.path.join(directory, name)
            self._read_config_from_file(path, encoding)

    def _read_config_from_file_require(self, path, encoding):
        assert aeidon.encodings.is_valid_code(encoding)

    def _read_config_from_file(self, path, encoding):
        """Read configurations from file at `path`."""
        if not os.path.isfile(path): return
        basename = os.path.basename(path)
        extension = ".%s.conf" % self.pattern_type
        code = basename.replace(extension, "")
        if not code in self._patterns: return
        patterns = self._patterns[code]
        for element in ET.parse(path).findall("pattern"):
            name = unicode(element.get("name"))
            name = name.replace("&quot;", '"')
            name = name.replace("&amp;", "&")
            enabled = (element.get("enabled") == "true")
            for pattern in patterns:
                if pattern.get_name(False) == name:
                    pattern.enabled = enabled

    def _read_patterns(self):
        """Read all patterns of :attr:`pattern_type` from files."""
        data_dir = os.path.join(aeidon.DATA_DIR, "patterns")
        data_home_dir = os.path.join(aeidon.DATA_HOME_DIR, "patterns")
        config_home_dir = os.path.join(aeidon.CONFIG_HOME_DIR, "patterns")
        encoding = aeidon.util.get_default_encoding()
        self._read_patterns_from_directory(data_dir, "utf_8")
        self._read_patterns_from_directory(data_home_dir, encoding)
        self._read_config_from_directory(data_dir, "utf_8")
        self._read_config_from_directory(config_home_dir, encoding)

    def _read_patterns_from_directory_require(self, directory, encoding):
        assert aeidon.encodings.is_valid_code(encoding)

    def _read_patterns_from_directory(self, directory, encoding):
        """Read all patterns from files in `directory`."""
        if not os.path.isdir(directory): return
        extension = ".%s" % self.pattern_type
        extensions = (extension, "%s.in" % extension)
        files = [x for x in os.listdir(directory)
                 if x.endswith(extensions)]

        for name in [x for x in files if x.endswith(".in")]:
            # If both untranslated and translated pattern files are found,
            # load patterns only from the translated one.
            if name[:-3] in files:
                files.remove(name)
        for name in files:
            path = os.path.join(directory, name)
            self._read_patterns_from_file(path, encoding)

    def _read_patterns_from_file_require(self, directory, encoding):
        assert aeidon.encodings.is_valid_code(encoding)

    def _read_patterns_from_file(self, path, encoding):
        """Read all patterns from file at `path`."""
        if not os.path.isfile(path): return
        basename = os.path.basename(path)
        extension = ".%s" % self.pattern_type
        if basename.endswith(".in"):
            extension = ".%s.in" % self.pattern_type
        code = basename.replace(extension, "")
        local = path.startswith(aeidon.DATA_HOME_DIR)
        patterns = self._patterns.setdefault(code, [])
        lines = aeidon.util.readlines(path, encoding)
        lines = [self._re_comment.sub("", x) for x in lines]
        lines = [x.strip() for x in lines]
        for line in (x for x in lines if x):
            if line.startswith("["): # [HEADER]
                patterns.append(aeidon.Pattern())
                patterns[-1].local = local
            else: # [_]KEY=VALUE
                name, value = unicode(line).split("=", 1)
                name = (name[1:] if name.startswith("_") else name)
                patterns[-1].set_field(name, value)

    def _write_config_to_file(self, code, encoding):
        """Write configurations of all patterns to file."""
        local_dir = os.path.join(aeidon.CONFIG_HOME_DIR, "patterns")
        if not os.path.isdir(local_dir): return
        basename = "%s.%s.conf" % (code, self.pattern_type)
        path = os.path.join(local_dir, basename)
        text = '<?xml version="1.0" encoding="utf-8"?>'
        text += '%s<patterns>%s' % (os.linesep, os.linesep)
        written_names = set(())
        for pattern in self._patterns[code]:
            name = pattern.get_name(False)
            if name in written_names: continue
            written_names.add(name)
            name = name.replace("&", "&amp;")
            name = name.replace('"', "&quot;")
            enabled = ("true" if pattern.enabled else "false")
            text += '  <pattern name="%s" ' % name
            text += 'enabled="%s"/>' % enabled
            text += os.linesep
        text += "</patterns>%s" % os.linesep
        aeidon.util.write(path, text, encoding)

    def get_countries(self, script, language):
        """Return a sequence of countries for which patterns exist."""
        codes = self._patterns.keys()
        start = "%s-%s-" % (script, language)
        codes = [x for x in codes if x.startswith(start)]
        countries = [x.split("-")[2] for x in codes]
        return tuple(aeidon.util.get_unique(countries))

    def get_languages(self, script):
        """Return a sequence of languages for which patterns exist."""
        codes = self._patterns.keys()
        start = "%s-" % script
        codes = [x for x in codes if x.startswith(start)]
        languages = [x.split("-")[1] for x in codes]
        return tuple(aeidon.util.get_unique(languages))

    def get_patterns_require(self, script=None, language=None, country=None):
        self._assert_identifiers(script, language, country)

    def get_patterns(self, script=None, language=None, country=None):
        """Return patterns for `script`, `language` and `country`."""
        patterns = []
        for code in self._get_codes(script, language, country):
            for pattern in self._patterns.get(code, []):
                patterns.append(pattern)
        patterns = self._filter_patterns(patterns)
        return patterns

    def get_scripts(self):
        """Return a sequence of scripts for which patterns exist."""
        codes = self._patterns.keys()
        while "Zyyy" in codes:
            codes.remove("Zyyy")
        scripts = [x.split("-")[0] for x in codes]
        return tuple(aeidon.util.get_unique(scripts))

    def save_config_require(self, script=None, language=None, country=None):
        self._assert_identifiers(script, language, country)

    def save_config(self, script=None, language=None, country=None):
        """Save pattern configurations to files."""
        local_dir = os.path.join(aeidon.CONFIG_HOME_DIR, "patterns")
        aeidon.deco.silent(OSError)(aeidon.util.makedirs)(local_dir)
        encoding = aeidon.util.get_default_encoding()
        codes = self._get_codes(script, language, country)
        for code in (x for x in codes if x in self._patterns):
            self._write_config_to_file(code, "utf_8")



# Copyright (C) 2005-2008,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import gaupol


class TestUpdateAgent(gaupol.TestCase):

    def setup_method(self, method):
        self.application = self.new_application()
        self.application.add_page(self.new_page())

    def test_flash_message(self):
        self.application.flash_message("")
        self.application.flash_message(None)

    def test_on_activate_next_project_activate(self):
        self.application.notebook.set_current_page(0)
        self.application.get_action("activate_next_project").activate()

    def test_on_activate_previous_project_activate(self):
        self.application.notebook.set_current_page(-1)
        self.application.get_action("activate_previous_project").activate()

    def test_on_conf_application_window_notify_toolbar_style(self):
        for style in gaupol.toolbar_styles:
            gaupol.conf.application_window.toolbar_style = style

    def test_on_move_tab_left_activate(self):
        self.application.notebook.set_current_page(-1)
        self.application.get_action("move_tab_left").activate()

    def test_on_move_tab_right_activate(self):
        self.application.notebook.set_current_page(0)
        self.application.get_action("move_tab_right").activate()

    def test_on_notebook_page_reordered(self):
        self.application.notebook.set_current_page(-1)
        self.application.get_action("move_tab_left").activate()

    def test_on_notebook_switch_page(self):
        self.application.notebook.set_current_page(0)
        self.application.notebook.set_current_page(-1)

    def test_on_view_move_cursor(self):
        page = self.application.get_current_page()
        page.view.emit("move-cursor", 1, 1)
        page.view.emit("move-cursor", 1, 1)
        gaupol.util.iterate_main()

    def test_on_view_selection_changed(self):
        page = self.application.get_current_page()
        page.view.select_rows((0,))
        page.view.select_rows((1,))

    def test_on_window_window_state_event(self):
        self.application.window.unmaximize()
        self.application.window.maximize()
        self.application.window.unmaximize()
        gaupol.util.iterate_main()

    def test_push_message(self):
        self.application.push_message("")
        self.application.push_message(None)

    def test_update_gui(self):
        self.application.update_gui()

    def test_update_gui__close_all(self):
        self.application.get_action("close_all_projects").activate()
        self.application.update_gui()

    def test_update_gui__revert(self):
        page = self.application.get_current_page()
        page.project.remove_subtitles((0,))
        page.project.remove_subtitles((0,))
        self.application.update_gui()
        page.project.undo()
        self.application.update_gui()



# Copyright (C) 2005-2008,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Building and updating dynamic menus."""

import aeidon
import gtk
_ = aeidon.i18n._


class MenuAgent(aeidon.Delegate):

    """Building and updating menus.

    :ivar _projects_id: :class:`gtk.UIManager` merge ID for projects menu
    :ivar _redo_menu_items: Redo menu tool button menu items
    :ivar _undo_menu_items: Undo menu tool button menu items
    """

    __metaclass__ = aeidon.Contractual

    def __init__(self, master):
        """Initialize a MenuAgent object."""
        aeidon.Delegate.__init__(self, master)
        self._projects_id = None
        self._redo_menu_items = []
        self._undo_menu_items = []

    def _add_project_action(self, page):
        """Add an action to the "projects" action group for `page`."""
        index = self.pages.index(page)
        basename = page.get_main_basename()
        name = "activate_project_%d" % index
        label = page.tab_label.get_text().replace("_", "__")
        label = "%d. %s" % (index + 1, label)
        label = ("_%s" % label if index < 9 else label)
        tooltip = _('Activate "%s"') % basename
        action = gtk.RadioAction(name, label, tooltip, None, index)
        action_group = self.get_action_group("projects")
        group = action_group.get_action("activate_project_0")
        if group is not None:
            action.set_group(group)
        accel = ("<alt>%d" % (index + 1) if index < 9 else None)
        action_group.add_action_with_accel(action, accel)
        action.connect("changed", self._on_projects_action_changed)
        action.set_active(page is self.get_current_page())
        return action.get_name()

    def _on_projects_action_changed(self, item, active_item):
        """Change the page in the notebook to the selected project."""
        index = int(active_item.get_name().split("_")[-1])
        self.notebook.set_current_page(index)

    @aeidon.deco.export
    def _on_redo_button_show_menu(self, *args):
        """Show a menu listing all redoable actions."""
        menu = gtk.Menu()
        self._redo_menu_items = []
        page = self.get_current_page()
        for i, action in enumerate(page.project.redoables):
            item = gtk.MenuItem(action.description, False)
            item.set_data("index", i)
            item.set_data("tooltip", _('Redo "%s"') % action.description)
            callback = self._on_redo_menu_item_activate
            item.connect("activate", callback)
            callback = self._on_redo_menu_item_enter_notify_event
            item.connect("enter-notify-event", callback)
            callback = self._on_redo_menu_item_leave_notify_event
            item.connect("leave-notify-event", callback)
            self._redo_menu_items.append(item)
            menu.append(item)
        menu.show_all()
        self.get_tool_item("redo_action").set_menu(menu)

    def _on_redo_menu_item_activate(self, menu_item):
        """Redo the selected action and all those above it."""
        index = menu_item.get_data("index")
        self.redo(index + 1)

    def _on_redo_menu_item_enter_notify_event(self, menu_item, event):
        """Show tooltip and select all actions above `menu_item`."""
        index = menu_item.get_data("index")
        for item in self._redo_menu_items[:index]:
            item.set_state(gtk.STATE_PRELIGHT)
        self.push_message(menu_item.get_data("tooltip"))

    def _on_redo_menu_item_leave_notify_event(self, menu_item, event):
        """Hide tooltip and unselect all actions above `menu_item`."""
        index = menu_item.get_data("index")
        for item in self._redo_menu_items[:index]:
            item.set_state(gtk.STATE_NORMAL)
        self.push_message(None)

    @aeidon.deco.export
    def _on_show_projects_menu_activate(self, *args):
        """Update all project actions in the projects menu."""
        action_group = self.get_action_group("projects")
        for action in action_group.list_actions():
            action_group.remove_action(action)
        if self._projects_id is not None:
            self.uim.remove_ui(self._projects_id)
        page = self.get_current_page()
        if page is None: return
        ui  = '<ui><menubar name="menubar">'
        ui += '<menu name="projects" action="show_projects_menu">'
        ui += '<placeholder name="open">'
        for i, page in enumerate(self.pages):
            name = self._add_project_action(page)
            ui += '<menuitem name="%d" action="%s"/>' % (i, name)
        ui += '</placeholder></menu></menubar></ui>'
        self._projects_id = self.uim.add_ui_from_string(ui)
        self.uim.ensure_update()
        self.set_menu_notify_events("projects")

    @aeidon.deco.export
    def _on_tab_widget_button_press_event(self, button, event, page):
        """Display a pop-up menu with tab-related actions."""
        if event.button != 3: return
        if page is not self.get_current_page():
            self.set_current_page(page)
        menu = self.uim.get_widget("/ui/tab_popup")
        menu.popup(None, None, None, event.button, event.time)

    @aeidon.deco.export
    def _on_undo_button_show_menu(self, *args):
        """Show a menu listing all undoable actions."""
        menu = gtk.Menu()
        self._undo_menu_items = []
        page = self.get_current_page()
        for i, action in enumerate(page.project.undoables):
            item = gtk.MenuItem(action.description, False)
            item.set_data("index", i)
            item.set_data("tooltip", _('Undo "%s"') % action.description)
            callback = self._on_undo_menu_item_activate
            item.connect("activate", callback)
            callback = self._on_undo_menu_item_enter_notify_event
            item.connect("enter-notify-event", callback)
            callback = self._on_undo_menu_item_leave_notify_event
            item.connect("leave-notify-event", callback)
            self._undo_menu_items.append(item)
            menu.append(item)
        menu.show_all()
        self.get_tool_item("undo_action").set_menu(menu)
    def _on_undo_menu_item_activate(self, menu_item):
        """Undo the selected action and all those above it."""
        index = menu_item.get_data("index")
        self.undo(index + 1)

    def _on_undo_menu_item_enter_notify_event(self, menu_item, event):
        """Show tooltip and select all actions above `menu_item`."""
        index = menu_item.get_data("index")
        for item in self._undo_menu_items[:index]:
            item.set_state(gtk.STATE_PRELIGHT)
        self.push_message(menu_item.get_data("tooltip"))

    def _on_undo_menu_item_leave_notify_event(self, menu_item, event):
        """Hide tooltip and unselect all actions above `menu_item`."""
        index = menu_item.get_data("index")
        for item in self._undo_menu_items[:index]:
            item.set_state(gtk.STATE_NORMAL)
        self.push_message(None)

    @aeidon.deco.export
    def set_menu_notify_events(self, name):
        """Set statusbar tooltips for menu items of action group."""
        def on_enter(menu_item, event, self, action):
            self.push_message(action.props.tooltip)
        def on_leave(menu_item, event, self, action):
            self.push_message(None)
        action_group = self.get_action_group(name)
        for action in action_group.list_actions():
            for widget in action.get_proxies():
                widget.connect("enter-notify-event", on_enter, self, action)
                widget.connect("leave-notify-event", on_leave, self, action)



# Copyright (C) 2007-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import aeidon


class PuppetBaseClass(object):

    __metaclass__ = aeidon.Contractual

    def _invariant(self):
        assert True

    def do_require(self):
        assert True

    def do_ensure(self, value):
        assert value is None

    def do(self):
        return None


class PuppetClassImplemented(PuppetBaseClass):

    def _invariant(self):
        assert True

    def do_require(self):
        assert True

    def do_ensure(self, value):
        assert value is None

    def do(self):
        return None


class PuppetClassNotImplemented(PuppetBaseClass):

    def do(self):
        return None


class TestContractual(aeidon.TestCase):

    def setup_method(self, method):
        self.contractual = PuppetClassImplemented()

    def test___new____implemented(self):
        self.contractual = PuppetClassImplemented()
        self.contractual.do()

    def test___new____not_implemented(self):
        self.contractual = PuppetClassNotImplemented()
        self.contractual.do()



# Copyright (C) 2005-2008,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Previewing subtitles with a video player."""

import aeidon
import gaupol
import glib
import gtk
import tempfile
_ = aeidon.i18n._


class PreviewAgent(aeidon.Delegate):

    """Previewing subtitles with a video player."""

    def _check_process_state(self, page, process, output_path, command):
        """Check if `process` has terminated or not."""
        if process.poll() is None: return True
        self._handle_output(process, output_path, command)
        return False # to not check again.

    def _handle_output(self, process, output_path, command):
        """Handle output of finished `process`."""
        with open(output_path, "r") as fobj:
            output = fobj.read()
        output = "$ %s\n\n%s" % (command, output)
        aeidon.temp.remove(output_path)
        self.output_window.set_output(output)
        if process.returncode == 0: return
        dialog = gaupol.PreviewErrorDialog(self.window, output)
        gaupol.util.flash_dialog(dialog)

    def _show_encoding_error_dialog(self):
        """Show an error dialog after failing to encode file."""
        title = _('Failed to encode subtitle file to temporary '
                  'directory "%s"') % tempfile.gettempdir()
        message = _("Subtitle data could not be encoded to a temporary file "
            "for preview with the current character encoding. Please first "
            "save the subtitle file with a different character encoding.")
        dialog = gaupol.ErrorDialog(self.window, title, message)
        dialog.add_button(gtk.STOCK_OK, gtk.RESPONSE_OK)
        gaupol.util.flash_dialog(dialog)

    def _show_io_error_dialog(self, message):
        """Show an error dialog after failing to write file."""
        title = _('Failed to save subtitle file to temporary '
                  'directory "%s"') % tempfile.gettempdir()
        dialog = gaupol.ErrorDialog(self.window, title, message)
        dialog.add_button(gtk.STOCK_OK, gtk.RESPONSE_OK)
        gaupol.util.flash_dialog(dialog)

    def _show_process_error_dialog(self, message):
        """Show an error dialog after failing to launch video player."""
        title = _("Failed to launch video player")
        dialog = gaupol.ErrorDialog(self.window, title, message)
        dialog.add_button(gtk.STOCK_OK, gtk.RESPONSE_OK)
        gaupol.util.flash_dialog(dialog)

    @aeidon.deco.export
    def _on_preview_activate(self, *args):
        """Preview from selected position with a video player."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        row = (rows[0] if rows else 0)
        position = page.project.subtitles[row].start
        col = page.view.get_focus()[1]
        if col == page.view.columns.TRAN_TEXT:
            doc = aeidon.documents.TRAN
        else: # Any other column previews the main file.
            doc = aeidon.documents.MAIN
        self.preview(page, position, doc)

    @aeidon.deco.export
    def preview(self, page, position, doc, temp=False):
        """Preview from `position` with a video player."""
        command = gaupol.util.get_preview_command()
        offset = gaupol.conf.preview.offset
        encoding = ("utf_8" if gaupol.conf.preview.force_utf_8 else None)
        try: process, command, output_path = page.project.preview(position,
                                                                  doc,
                                                                  command,
                                                                  offset,
                                                                  encoding,
                                                                  temp)

        except aeidon.ProcessError as message:
            return self._show_process_error_dialog(message)
        except (IOError, OSError) as (no, message):
            return self._show_io_error_dialog(message)
        except UnicodeError:
            return self._show_encoding_error_dialog()
        # glib.child_watch_add does not appear to work on Windows,
        # so let's watch the process by polling it at regular intervals.
        glib.timeout_add(1000,
                         self._check_process_state,
                         page,
                         process,
                         output_path,
                         command)

    @aeidon.deco.export
    def preview_changes(self, page, row, doc, method, args=None, kwargs=None):
        """Preview changes caused by `method` with a video player."""
        subtitles = [x.copy() for x in page.project.subtitles]
        framerate = page.project.framerate
        blocked = page.project.block_all()
        method(register=None, *(args or ()), **(kwargs or {}))
        position = page.project.subtitles[row].start
        page.project.unblock_all(blocked)
        self.preview(page, position, doc, temp=True)
        page.project.set_framerate(framerate, register=None)
        page.project.subtitles = subtitles



# Copyright (C) 2006-2008,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import gaupol
import gtk


class TestModule(gaupol.TestCase):

    def run__text_view(self):
        text_view = gtk.TextView()
        text_buffer = text_view.get_buffer()
        text = ("Everything has been said\n"
                "provided words do not change\n"
                "their meanings\n"
                "and meanings their words")

        text_buffer.insert_at_cursor(text)
        gaupol.ruler.connect_text_view(text_view)
        scroller = gtk.ScrolledWindow()
        scroller.set_shadow_type(gtk.SHADOW_IN)
        scroller.add(text_view)
        window = gtk.Window()
        window.connect("delete-event", gtk.main_quit)
        window.set_position(gtk.WIN_POS_CENTER)
        window.set_default_size(300, 100)
        window.set_border_width(12)
        window.add(scroller)
        window.show_all()
        gtk.main()

    def test_connect_text_view(self):
        text_view = gtk.TextView()
        gaupol.ruler.connect_text_view(text_view)
        text_buffer = text_view.get_buffer()
        text_buffer.insert_at_cursor("test\ntest")

    def test_disconnect_text_view(self):
        text_view = gtk.TextView()
        gaupol.ruler.connect_text_view(text_view)
        gaupol.ruler.disconnect_text_view(text_view)
        text_buffer = text_view.get_buffer()
        text_buffer.insert_at_cursor("test\ntest")

    def test_get_length_function__char(self):
        unit = gaupol.length_units.CHAR
        function = gaupol.ruler.get_length_function(unit)
        assert function("iii") == 3

    def test_get_length_function__em(self):
        unit = gaupol.length_units.EM
        function = gaupol.ruler.get_length_function(unit)
        assert 0.5 < function("M") < 1.5

    def test_get_lengths__char(self):
        unit = gaupol.length_units.CHAR
        gaupol.conf.editor.length_unit = unit
        lengths = gaupol.ruler.get_lengths("MMM\n<i>iii</i>")
        assert tuple(lengths) == (3, 3)

    def test_get_lengths__em(self):
        unit = gaupol.length_units.EM
        gaupol.conf.editor.length_unit = unit
        gaupol.conf.editor.use_custom_font = False
        lengths = gaupol.ruler.get_lengths("MMM\n<i>i</i>")
        assert tuple(lengths) == (3, 0)



# Copyright (C) 2005-2008,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Dialog for informing that preview failed."""

import gaupol
import gtk


class PreviewErrorDialog(gaupol.BuilderDialog):

    """Dialog for informing that preview failed."""

    _widgets = ("text_view",)

    def __init__(self, parent, output):
        """Initialize a :class:`PreviewErrorDialog` object."""
        gaupol.BuilderDialog.__init__(self, "previewerr-dialog.ui")
        self._init_data(output)
        gaupol.util.scale_to_content(self._text_view,
                                     min_nlines=5,
                                     max_nchar=100,
                                     max_nlines=30,
                                     font="monospace")

        self._dialog.set_transient_for(parent)
        self._dialog.set_default_response(gtk.RESPONSE_OK)

    def _init_data(self, output):
        """Initialize output text in the text view."""
        text_buffer = self._text_view.get_buffer()
        text_buffer.create_tag("output", family="monospace")
        itr = text_buffer.get_end_iter()
        text_buffer.insert_with_tags_by_name(itr, output, "output")



# Copyright (C) 2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Reading, writing and storing configurations."""

import aeidon
import codecs
import gaupol
import json
import os
import sys

__all__ = ("ConfigurationStore",)


config_defaults = {
    "application_window": {
        "maximized": False,
        "position": [0, 0],
        "show_main_toolbar": True,
        "show_statusbar": True,
        "show_video_toolbar": False,
        "size": [700, 433],
        "toolbar_style": gaupol.toolbar_styles.DEFAULT,
        },
    "capitalization": {
        "country": "",
        "language": "",
        "script": "Latn",
        },
    "common_error": {
        "classes": ["Human", "OCR"],
        "country": "",
        "language": "",
        "script": "Latn",
        },
    "debug": {
        "text_editor": "gedit +$LINENO $FILE",
        },
    "duration_adjust": {
        "gap": 0.0,
        "lengthen": True,
        "maximum": 6.0,
        "minimum": 1.5,
        "speed": 15.0,
        "shorten": False,
        "target": gaupol.targets.CURRENT,
        "use_minimum": True,
        "use_gap": True,
        "use_maximum": False,
        },
    "general": {
        "version": None,
        },
    "editor": {
        "custom_font": "",
        "field_order": [gaupol.fields.NUMBER,
                        gaupol.fields.START,
                        gaupol.fields.END,
                        gaupol.fields.DURATION,
                        gaupol.fields.MAIN_TEXT,
                        gaupol.fields.TRAN_TEXT],

        "framerate": aeidon.framerates.FPS_23_976,
        "length_unit": gaupol.length_units.EM,
        "mode": aeidon.modes.TIME,
        "show_lengths_cell": True,
        "show_lengths_edit": True,
        "undo_limit": 50,
        "use_custom_font": False,
        "use_undo_limit": False,
        "visible_fields": [gaupol.fields.NUMBER,
                           gaupol.fields.START,
                           gaupol.fields.END,
                           gaupol.fields.DURATION,
                           gaupol.fields.MAIN_TEXT],
        },
    "encoding": {
        "fallback": ["utf_8", "cp1252"],
        "try_auto": True,
        "try_locale": True,
        "visible": ["utf_8", "cp1252"],
        },
    "extensions": {
        "active": [],
        },
    "file": {
        "align_method": aeidon.align_methods.POSITION,
        "directory": "",
        "encoding": aeidon.util.get_default_encoding() or "utf_8",
        "format": aeidon.formats.SUBRIP,
        "max_recent": 10,
        "newline": aeidon.util.get_default_newline(),
        },
    "framerate_convert": {
        "target": gaupol.targets.CURRENT,
        },
    "hearing_impaired": {
        "country": "",
        "language": "",
        "script": "Latn",
        },
    "join_split_words": {
        "join": True,
        "split": False,
        },
    "line_break": {
        "country": "",
        "language": "",
        "length_unit": gaupol.length_units.EM,
        "max_deviation": 0.16,
        "max_length": 28,
        "max_lines": 2,
        "script": "Latn",
        "skip_max_length": 28,
        "skip_max_lines": 3,
        "use_skip_max_length": True,
        "use_skip_max_lines": True,
        },
    "output_window": {
        "maximized": False,
        "position": [0, 0],
        "show": False,
        "size": [600, 371],
        },
    "position_shift": {
        "target": gaupol.targets.CURRENT,
        },
    "position_transform": {
        "target": gaupol.targets.CURRENT,
        },
    "preview": {
        "custom_command": "",
        "force_utf_8": True,
        "offset": 5.0,
        "player": (aeidon.players.VLC if sys.platform == "win32"
                   else aeidon.players.MPLAYER),

        "use_custom_command": False,
        },
    "search": {
        "fields": [gaupol.fields.MAIN_TEXT],
        "ignore_case": True,
        "max_history": 10,
        "regex": False,
        "target": gaupol.targets.CURRENT,
        },
    "speech_recognition": {
        "acoustic_model": "",
        "advance_length": 100,
        "lang_model": "",
        "noise_level": 256,
        "phonetic_dict": "",
        "silence_length": 300,
        "use_custom_models": False,
        },
    "spell_check": {
        "field": gaupol.fields.MAIN_TEXT,
        "inline": False,
        "language": aeidon.locales.get_system_code() or "en",
        "target": gaupol.targets.CURRENT,
        },
    "subtitle_insert": {
        "above": False,
        },
    "text_assistant": {
        "field": gaupol.fields.MAIN_TEXT,
        "maximized": False,
        "pages": [],
        "remove_blank": True,
        "size": [700, 433],
        "target": gaupol.targets.CURRENT,
        },
    }

config_enums = {
    "application_window": {
        "toolbar_style": gaupol.toolbar_styles,
        },
    "duration_adjust": {
        "target": gaupol.targets,
        },
    "editor": {
        "field_order": gaupol.fields,
        "framerate": aeidon.framerates,
        "length_unit": gaupol.length_units,
        "mode": aeidon.modes,
        "visible_fields": gaupol.fields,
        },
    "extensions": {
        },
    "file": {
        "align_method": aeidon.align_methods,
        "format": aeidon.formats,
        "newline": aeidon.newlines,
        },
    "framerate_convert": {
        "target": gaupol.targets,
        },
    "line_break": {
        "length_unit": gaupol.length_units,
        },
    "position_shift": {
        "target": gaupol.targets,
        },
    "position_transform": {
        "target": gaupol.targets,
        },
    "preview": {
        "player": aeidon.players,
        },
    "search": {
        "fields": gaupol.fields,
        "target": gaupol.targets,
        },
    "spell_check": {
        "field": gaupol.fields,
        "target": gaupol.targets,
        },
    "text_assistant": {
        "field": gaupol.fields,
        "target": gaupol.targets,
        },
    }


class EnumDecoder(json.JSONDecoder):

    """JSON decoder for enumerations of :mod:`aeidon` and :mod:`gaupol`."""

    def __init__(self, *args, **kwargs):
        """Initialize a :class:`EnumDecoder` object.

        `kwargs` should contain an "enum" key with a value of ``None`` if not
        decoding an enumeration or the corresponding
        :class:`aeidon.Enumeration` instance if decoding an enumeration.
        """
        self.enum = kwargs["enum"]
        del kwargs["enum"]
        json.JSONDecoder.__init__(self, *args, **kwargs)

    # pylint: disable=W0221
    def decode(self, string):
        """Return Python object matching JSON `string`."""
        if self.enum is None:
            return json.JSONDecoder.decode(self, string)
        if string.startswith("[") and string.endswith("]"):
            return [getattr(self.enum, x.strip())
                    for x in string[1:-1].split(",")]

        return getattr(self.enum, string)


class ConfigurationStore(gaupol.AttributeDictionary):

    """Reading, writing and storing configurations.

    :cvar _defaults: Dictionary of default values of options
    :cvar _enums: Dictionary of :class:`aeidon.Enumeration` instances
    :ivar path: Path to user's local configuration file
    """

    _defaults = config_defaults
    _enums = config_enums

    def __init__(self):
        """Initialize a :class:`ConfigurationStore` object."""
        root = aeidon.util.copy_dict(self._defaults)
        gaupol.AttributeDictionary.__init__(self, root)
        self.path = None

    def _flatten(self, values):
        """Return a flattened version of `values` dictionary."""
        # pylint: disable=W0631
        def flatten(deep, parent):
            flat_dict = {parent: {}}
            deep = aeidon.util.copy_dict(deep)
            for key, value in deep.iteritems():
                if isinstance(value, dict):
                    key = "::".join((parent, key))
                    value = flatten(value, key)
                    flat_dict.update(value)
                else: # Non-dictionary key.
                    flat_dict[parent][key] = value
            if not flat_dict[parent]:
                del flat_dict[parent]
            return flat_dict
        final_dict = {}
        values = aeidon.util.copy_dict(values)
        for key, value in values.iteritems():
            if isinstance(value, dict):
                final_dict.update(flatten(value, key))
            else: # Non-dictionary key.
                final_dict[key] = value
        return final_dict

    def connect_notify(self, sections, option, obj, *args):
        """Connect `option`'s notify signal to `obj`'s callback method."""
        if isinstance(sections, basestring):
            sections = (sections,)
        container = self
        for section in sections:
            container = getattr(container, section)
        signal = "notify::%s" % option
        method_name = "_on_conf_%s_%s" % ("_".join(sections),
                                          signal.replace("::", "_"))

        if not hasattr(obj, method_name):
            method_name = method_name[1:]
        method = getattr(obj, method_name)
        container.connect(signal, method, *args)

    def disconnect_notify(self, sections, option, obj):
        """Disconnect `option`'s notify signal from `obj`'s callback method."""
        if isinstance(sections, basestring):
            sections = (sections,)
        container = self
        for section in sections:
            container = getattr(container, section)
        signal = "notify::%s" % option
        method_name = "_on_conf_%s_%s" % ("_".join(sections),
                                          signal.replace("::", "_"))

        if not hasattr(obj, method_name):
            method_name = method_name[1:]
        method = getattr(obj, method_name)
        container.disconnect(signal, method)

    def query_default(self, sections, option):
        """Return default value of configuration option.

        Raise :exc:`KeyError` if section or option not found.
        """
        if isinstance(sections, basestring):
            sections = (sections,)
        container = self._defaults
        for section in sections:
            container = container[section]
        return container[option]

    def read_from_file(self):
        """Read values of configuration options from file.

        Raise :exc:`IOError` if reading file fails.
        Raise :exc:`UnicodeError` if decoding file fails.
        Fail silently if :attr:`path` is not set.
        """
        if self.path is None: return
        if not os.path.isfile(self.path): return
        encoding = aeidon.util.get_default_encoding()
        try:
            # Ignore all decoding errors, since all keys and all standard
            # values are all ASCII. This will only mangle recent etc.
            # filenames, which are always checked for existance anyway.
            lines = codecs.open(self.path,
                                "r",
                                encoding,
                                errors="ignore").readlines()

        except IOError:
            aeidon.util.print_read_io(sys.exc_info(), self.path)
            raise # IOError
        except UnicodeError:
            aeidon.util.print_read_unicode(sys.exc_info(), self.path, encoding)
            raise # UnicodeError
        lines = map(lambda x: x.strip(), lines)
        lines = filter(lambda x: x and not x.startswith("#"), lines)
        for line in lines:
            if line.startswith("[") and line.endswith("]"):
                sections = line[1:-1].strip().split("::")
                container = self
                enums = self._enums
                for section in sections:
                    if not hasattr(container, section):
                        container.add_attribute(section, {})
                    container = getattr(container, section)
                    enums = enums.get(section, {})
            else: # OPTION = VALUE
                option, value = line.split("=", 1)
                option = option.strip()
                value = value.strip()
                enum = enums.get(option, None)
                try: value = json.loads(value, cls=EnumDecoder, enum=enum)
                except (AttributeError, ValueError):
                    print ("Failed to parse value '%s' of option '%s.%s' "
                           "from configuration file '%s'." % (
                            value, "::".join(sections), option, self.path))

                    continue
                if hasattr(container, option) and value is None:
                    # Discard a None-value for all non-extension options.
                    # None-values are not used for any options, but might
                    # accidentally bet set in some corner or error cases.
                    # By discarding them here, we ensure a clean start.
                    print ("Discarding value '%s' of option '%s.%s' "
                           "from configuration file '%s'." % (
                            value, "::".join(sections), option, self.path))

                    continue
                if not hasattr(container, option):
                    # Add attribute if it does not exist in container, which
                    # has been initialized from config_defaults. This is needed
                    # for extensions, defaults of which are not known until
                    # later. Any possible obsolete options created here will be
                    # removed before writing in write_to_file.
                    container.add_attribute(option, value)
                setattr(container, option, value)

    def register_extension(self, name, defaults, enums=None):
        """Add section and options for extension.

        `name` should be preferrably the module name of the extension and it
        will appear in the section name as ``extensions::name``. `defaults`
        should be a dictionary of default values for options. `enums` should be
        a dictionary of :class:`aeidon.Enumeration` instances corresponding to
        enumeration items that appear in options.
        """
        self._defaults["extensions"].update({name: defaults})
        self.extensions.extend({name: aeidon.util.copy_dict(defaults)})
        self._enums["extensions"].update({name: enums or {}})

    def restore_defaults(self):
        """Set all configuration options to their default values."""
        self.update(self._defaults)

    def write_to_file(self):
        """Write values of configuration options to file.

        Raise :exc:`IOError` or :exc:`OSError` if unable to create
        configuration directory or unable to write configuration file.
        Fail silently if :attr:`path` is not set.
        """
        if self.path is None: return
        self.general.version = gaupol.__version__
        if not os.path.isdir(os.path.dirname(self.path)):
            try: aeidon.util.makedirs(os.path.dirname(self.path))
            except (IOError, OSError):
                aeidon.util.print_write_io(sys.exc_info(), self.path)
                raise # IOError, OSError
        encoding = aeidon.util.get_default_encoding()
        try: fobj = codecs.open(self.path, "w", encoding)
        except (IOError, OSError):
            aeidon.util.print_write_io(sys.exc_info(), self.path)
            raise # IOError, OSError
        root = self._flatten(self._root)
        defaults = self._flatten(self._defaults)
        for section in sorted(root):
            fobj.write(os.linesep)
            fobj.write("[%s]" % section)
            fobj.write(os.linesep)
            for option in sorted(root[section]):
                value = root[section][option]
                json_value = json.dumps(value, ensure_ascii=False)
                if (not section.startswith("extensions::") and
                    (not section in defaults or
                     not option in defaults[section])):
                    # Discard removed options, but always keep
                    # all options of all extensions.
                    continue
                if (section in defaults) and (option in defaults[section]):
                    if value == defaults[section][option]:
                        fobj.write("# ")
                try: fobj.write("%s = %s" % (option, json_value))
                except UnicodeError:
                    print ("Failed to write value '%s' of option '%s.%s' "
                           "to configuration file '%s'." % (
                            value, section, option, self.path))

                fobj.write(os.linesep)



# Copyright (C) 2005-2008,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Dialogs for selecting character encodings."""

import aeidon
import gaupol
import gtk
_ = aeidon.i18n._

__all__ = ("EncodingDialog", "MenuEncodingDialog")


class EncodingDialog(gaupol.BuilderDialog):

    """Dialog for selecting a character encoding."""

    __metaclass__ = aeidon.Contractual

    _widgets = ("tree_view",)

    def __init__(self, parent):
        """Initialize an :class:`EncodingDialog` object."""
        gaupol.BuilderDialog.__init__(self, "encoding-dialog.ui")
        self._init_tree_view()
        gaupol.util.scale_to_content(self._tree_view,
                                     min_nchar=10,
                                     min_nlines=5,
                                     max_nchar=100,
                                     max_nlines=30)

        self._dialog.set_transient_for(parent)
        self._dialog.set_default_response(gtk.RESPONSE_OK)

    def _init_tree_view(self):
        """Initialize the tree view."""
        selection = self._tree_view.get_selection()
        selection.set_mode(gtk.SELECTION_SINGLE)
        store = gtk.ListStore(str, str, str)
        for item in aeidon.encodings.get_valid():
            store.append((item[0], item[2], item[1]))
        store.set_sort_column_id(1, gtk.SORT_ASCENDING)
        self._tree_view.set_model(store)
        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(_("Description"), renderer, text=1)
        column.set_clickable(True)
        column.set_sort_column_id(1)
        self._tree_view.append_column(column)
        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(_("Encoding"), renderer, text=2)
        column.set_clickable(True)
        column.set_sort_column_id(2)
        self._tree_view.append_column(column)

    def _on_tree_view_row_activated(self, *args):
        """Send response to select activated character encoding."""
        print "ACTIVATED"
        self.response(gtk.RESPONSE_OK)

    def get_encoding_ensure(self, value):
        if value is not None:
            assert aeidon.encodings.is_valid_code(value)

    def get_encoding(self):
        """Return the selected encoding or ``None``."""
        selection = self._tree_view.get_selection()
        store, itr = selection.get_selected()
        if itr is None: return
        return store.get_value(itr, 0)


class MenuEncodingDialog(EncodingDialog):

    """Dialog for selecting character encodings."""

    def _init_tree_view(self):
        """Initialize the tree view."""
        selection = self._tree_view.get_selection()
        selection.set_mode(gtk.SELECTION_SINGLE)
        store = gtk.ListStore(str, str, str, bool)
        visible = gaupol.conf.encoding.visible
        for item in aeidon.encodings.get_valid():
            store.append((item[0], item[2], item[1], item[0] in visible))
        store.set_sort_column_id(1, gtk.SORT_ASCENDING)
        self._tree_view.set_model(store)
        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(_("Description"), renderer, text=1)
        column.set_clickable(True)
        column.set_sort_column_id(1)
        self._tree_view.append_column(column)
        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(_("Encoding"), renderer, text=2)
        column.set_clickable(True)
        column.set_sort_column_id(2)
        self._tree_view.append_column(column)
        renderer = gtk.CellRendererToggle()
        renderer.connect("toggled", self._on_tree_view_cell_toggled)
        column = gtk.TreeViewColumn(_("Show in Menu"), renderer, active=3)
        column.set_sort_column_id(3)
        self._tree_view.append_column(column)

    def _on_tree_view_cell_toggled(self, renderer, row):
        """Toggle the value of the "Show in Menu" column."""
        store = self._tree_view.get_model()
        store[row][3] = not store[row][3]

    def get_visible_encodings_ensure(self, value):
        for name in value:
            assert aeidon.encodings.is_valid_code(name)

    def get_visible_encodings(self):
        """Return encodings chosen to be visible."""
        store = self._tree_view.get_model()
        return [store[i][0] for i in range(len(store)) if store[i][3]]



# Copyright (C) 2005-2008,2010-2011 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Base classes for :class:`gtk.UIManager` actions."""

import aeidon
import gaupol
import gtk
import sys

__all__ = ("Action",
           "MenuAction",
           "RadioAction",
           "RecentAction",
           "ToggleAction",
           "TopMenuAction",
           )


class Action(gtk.Action):

    """Base classes for :class:`gtk.UIManager` actions.

    :ivar accelerator: Accelerator string or ``None``

       :attr:`accelerator` defines a string in the format understood by the
       :func:`gtk.accelerator_parse`, ``None`` to use the stock accelerator or
       undefined to use a blank string.

    :ivar action_group: Name of action group to place action into

       Use "main-safe" for actions that do not conflict with widgets' own
       built-in keybindings (e.g. clipboard keybindings in entries) and
       "main-unsafe" for ones that do.

    :ivar widgets: Tuple of names of related application widgets

       :attr:`widgets` should be acquirable with :func:`getattr` from
       :class:`gaupol.Application`. Their sensitivities will be kep in sync.
    """

    accelerator = ""
    action_group = "main-unsafe"
    widgets = ()

    def __init__(self, name):
        """Initialize an :class:`Action` object."""
        gtk.Action.__init__(self, name, None, None, None)

    def _affirm_doable(self, application, page):
        """Raise :exc:`aeidon.AffirmationError` if action cannot be done."""
        pass

    def finalize(self, application):
        """Connect action to widgets and methods of `application`."""
        self.widgets = tuple(getattr(application, x) for x in self.widgets)
        callback = "_on_%s_activate" % self.props.name
        self.connect("activate", getattr(application, callback))

    def set_sensitive(self, sensitive):
        """Set the sensitivity of action and all its widgets."""
        for widget in self.widgets:
            widget.set_sensitive(sensitive)
        return gtk.Action.set_sensitive(self, sensitive)

    def update_sensitivity(self, application, page):
        """Update the sensitivity of action and all its widgets."""
        try:
            self._affirm_doable(application, page)
        except aeidon.AffirmationError:
            return self.set_sensitive(False)
        return self.set_sensitive(True)


class MenuAction(Action):

    """Base class for actions that are menu items with a submenu."""

    def finalize(self, application):
        """Connect action to widgets and methods of `application`."""
        self.widgets = tuple(getattr(application, x) for x in self.widgets)
        callback = "_on_%s_activate" % self.props.name
        if hasattr(application, callback):
            self.connect("activate", getattr(application, callback))


class RecentAction(gtk.RecentAction, Action):

    """Base class for :class:`gtk.UIManager` recent file actions.

    :ivar group: Name of :class:`gtk.RecentFilter` group
    """

    group = NotImplementedError

    def __init__(self, name):
        """Initialize an :class:`RecentAction` object."""
        gtk.RecentAction.__init__(self, name, None, None, None)
        if sys.platform == "win32":
            self.set_show_icons(False)
        self.set_show_numbers(False)
        self.set_show_not_found(False)
        self.set_show_tips(True)
        self.set_sort_type(gtk.RECENT_SORT_MRU)
        recent_filter = gtk.RecentFilter()
        recent_filter.add_group(self.group)
        self.add_filter(recent_filter)
        self.set_filter(recent_filter)
        self.set_data("group", self.group)
        self.set_limit(gaupol.conf.file.max_recent)

    def finalize(self, application):
        """Connect action to widgets of `application`."""
        self.widgets = tuple(getattr(application, x) for x in self.widgets)
        callback = "_on_%s_activate" % self.props.name
        if hasattr(application, callback):
            self.connect("activate", getattr(application, callback))
        callback = "_on_%s_item_activated" % self.props.name
        self.connect("item-activated", getattr(application, callback))


class ToggleAction(gtk.ToggleAction, Action):

    """Base class for UI manager toggle actions."""

    def __init__(self, name):
        """Initialize an :class:`ToggleAction` object."""
        gtk.ToggleAction.__init__(self, name, None, None, None)

    def finalize(self, application):
        """Connect action to widgets and methods of `application`."""
        self.widgets = tuple(getattr(application, x) for x in self.widgets)
        callback = "_on_%s_toggled" % self.props.name
        self.connect("toggled", getattr(application, callback))


class TopMenuAction(MenuAction):

    """Base class for actions that are top-level menu items with a submenu."""

    def update_sensitivity(self, application, page):
        """Update the sensitivity of action and all its widgets."""
        pass


class RadioAction(gtk.RadioAction, Action):

    """Base class for :class:`gtk.UIManager` radio actions.

    :ivar group: Class name of one action in the radio group
    """

    group = NotImplementedError

    def __init__(self, name):
        """Initialize an :class:`RadioAction` object."""
        gtk.RadioAction.__init__(self, name, None, None, None, 0)

    def finalize(self, application):
        """Connect action to widgets and methods of `application`."""
        self.widgets = tuple(getattr(application, x) for x in self.widgets)
        if self.__class__.__name__ != self.group: return
        callback = "_on_%s_changed" % self.props.name
        self.connect("changed", getattr(application, callback))



# Copyright (C) 2007-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Regular expression substitution for subtitle text."""

import aeidon
import re

__all__ = ("Pattern",)


class Pattern(aeidon.MetadataItem):

    """Regular expression substitution for subtitle text.

    :ivar enabled: ``True`` if pattern should be used, ``False`` if not
    :ivar fields: Dictionary of all data field names and values
    :ivar local: ``True`` if pattern is defined by user, ``False`` if system
    """

    def __init__(self, fields=None):
        """Initialize a :class:`Pattern` object."""
        aeidon.MetadataItem.__init__(self, fields)
        self.enabled = True
        self.local = False

    def get_flags(self):
        """Return the evaluated value of the ``Flags`` field."""
        flags = 0
        for name in self.get_field_list("Flags"):
            flags = flags | getattr(re, name)
        return flags



# Copyright (C) 2005-2008,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Cell renderer for multiline text data."""

import aeidon
import gaupol
import glib
import gtk

__all__ = ("MultilineCellRenderer",)


class CellTextView(gtk.TextView, gtk.CellEditable):

    """A :class:`gtk.TextView` suitable for cell renderer use."""

    __gtype_name__ = "CellTextView"

    def __init__(self, text_buffer=None):
        """Initialize a :class:`CellTextView` object."""
        gtk.TextView.__init__(self, text_buffer)
        gaupol.util.prepare_text_view(self)

    def do_editing_done(self, *args):
        """End editing."""
        self.remove_widget()

    def do_remove_widget(self, *args):
        """Remove widget."""
        pass

    def do_start_editing(self, *args):
        """Start editing."""
        pass

    def get_text(self):
        """Return text."""
        text_buffer = self.get_buffer()
        bounds = text_buffer.get_bounds()
        return text_buffer.get_text(*bounds)

    def set_text(self, text):
        """Set text."""
        self.get_buffer().set_text(text)


class MultilineCellRenderer(gtk.CellRendererText):

    """Cell renderer for multiline text data.

    If :attr:`gaupol.conf.editor.show_lengths_cell` is ``True``, line lengths
    are shown as superscripts at the end of each line.
    """

    __gtype_name__ = "MultilineCellRenderer"

    def __init__(self):
        """Initialize a MultilineCellRenderer object."""
        gtk.CellRendererText.__init__(self)
        self._in_editor_menu = False
        self._show_lengths = gaupol.conf.editor.show_lengths_cell
        self._text = ""
        gaupol.conf.connect_notify("editor", "show_lengths_cell", self)
        aeidon.util.connect(self, self, "notify::text")

    def _on_conf_editor_notify_show_lengths_cell(self, *args):
        """Hide or show line lengths if ``conf`` changed."""
        self._show_lengths = gaupol.conf.editor.show_lengths_cell

    def _on_editor_focus_out_event(self, editor, *args):
        """End editing."""
        if self._in_editor_menu: return
        editor.remove_widget()
        self.emit("editing-canceled")

    def _on_editor_key_press_event(self, editor, event):
        """End editing if ``Enter`` or ``Escape`` pressed."""
        if event.state & (gtk.gdk.SHIFT_MASK | gtk.gdk.CONTROL_MASK): return
        if event.keyval in (gtk.keysyms.Return, gtk.keysyms.KP_Enter):
            editor.remove_widget()
            self.emit("edited", editor.get_data("path"), editor.get_text())
        if event.keyval == gtk.keysyms.Escape:
            editor.remove_widget()
            self.emit("editing-canceled")

    def _on_editor_populate_popup(self, editor, menu):
        """Disable "focus-out-event" ending editing."""
        self._in_editor_menu = True
        def on_menu_unmap(menu, self):
            self._in_editor_menu = False
        menu.connect("unmap", on_menu_unmap, self)

    def _on_notify_text(self, *args):
        """Set markup by adding line lengths to text."""
        self._text = text = unicode(self.props.text)
        if not (text and self._show_lengths): return
        lengths = gaupol.ruler.get_lengths(text)
        text = glib.markup_escape_text(text)
        lines = text.split("\n")
        for i, line in filter(lambda x: x[1], enumerate(lines)):
            lines[i] += " <sup>%d</sup>" % lengths[i]
        self.props.markup = "\n".join(lines)

    def do_start_editing(self, event, widget, path, bg_area, cell_area, flags):
        """Initialize and return a :class:`CellTextView` widget."""
        editor = CellTextView()
        editor.modify_font(self.props.font_desc)
        editor.set_text(self._text)
        editor.set_size_request(cell_area.width, cell_area.height)
        editor.set_border_width(min(self.props.xpad, self.props.ypad))
        editor.set_data("path", path)
        editor.connect("focus-out-event", self._on_editor_focus_out_event)
        editor.connect("key-press-event", self._on_editor_key_press_event)
        editor.connect("populate-popup", self._on_editor_populate_popup)
        editor.show()
        return editor

    def set_show_lengths(self, show_lengths):
        """Show or hide line lengths overriding ``conf``."""
        self._show_lengths = show_lengths
        gaupol.conf.disconnect_notify("editor", "show_lengths_cell", self)



# Copyright (C) 2005-2008 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import gaupol
import gtk


class TestMultilineCellRenderer(gaupol.TestCase):

    def run__renderer(self):
        tree_view = gtk.TreeView()
        tree_view.set_headers_visible(False)
        store = gtk.ListStore(str)
        store.append(("test\ntest test",))
        tree_view.set_model(store)
        self.renderer.props.editable = True
        column = gtk.TreeViewColumn("", self.renderer, text=0)
        tree_view.append_column(column)
        window = gtk.Window()
        window.connect("delete-event", gtk.main_quit)
        window.set_position(gtk.WIN_POS_CENTER)
        window.set_default_size(240, 70)
        window.add(tree_view)
        window.show_all()
        gtk.main()

    def setup_method(self, method):
        self.renderer = gaupol.MultilineCellRenderer()



# Copyright (C) 2005-2008,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import aeidon
import gaupol
import gtk
import os


class TestCloseAgent(gaupol.TestCase):

    def run__confirm_close_main(self):
        page = self.application.get_current_page()
        page.project.remove_subtitles((0,))
        self.delegate._confirm_close_main(page)

    def run__confirm_close_translation(self):
        page = self.application.get_current_page()
        page.project.remove_subtitles((0,))
        self.delegate._confirm_close_translation(page)

    def setup_method(self, method):
        self.application = self.new_application()
        self.delegate = self.application.close.im_self

    def test__save_window_geometry(self):
        self.delegate._save_window_geometry()

    @aeidon.deco.monkey_patch(gaupol.util, "flash_dialog")
    @aeidon.deco.silent(gaupol.Default)
    def test_close__both_changed(self):
        gaupol.util.flash_dialog = lambda *args: gtk.RESPONSE_CANCEL
        self.application.open_main(self.new_subrip_file())
        self.application.open_translation(self.new_microdvd_file())
        self.application.pages[-1].project.remove_subtitles((0,))
        self.application.close(self.application.pages[-1], True)

    @aeidon.deco.monkey_patch(gaupol.util, "flash_dialog")
    @aeidon.deco.silent(gaupol.Default)
    def test_close__main_changed__discard(self):
        gaupol.util.flash_dialog = lambda *args: gtk.RESPONSE_NO
        doc = aeidon.documents.MAIN
        self.application.open_main(self.new_subrip_file())
        self.application.open_translation(self.new_microdvd_file())
        self.application.pages[-1].project.clear_texts((0,), doc)
        self.application.close(self.application.pages[-1], True)

    @aeidon.deco.monkey_patch(gaupol.util, "flash_dialog")
    @aeidon.deco.silent(gaupol.Default)
    def test_close__main_changed__save(self):
        gaupol.util.flash_dialog = lambda *args: gtk.RESPONSE_YES
        doc = aeidon.documents.MAIN
        self.application.open_main(self.new_subrip_file())
        self.application.open_translation(self.new_microdvd_file())
        self.application.pages[-1].project.clear_texts((0,), doc)
        self.application.close(self.application.pages[-1], True)

    @aeidon.deco.monkey_patch(gaupol.util, "flash_dialog")
    @aeidon.deco.silent(gaupol.Default)
    def test_close__main_removed(self):
        gaupol.util.flash_dialog = lambda *args: gtk.RESPONSE_CANCEL
        page = self.application.get_current_page()
        os.remove(page.project.main_file.path)
        self.application.close(page, True)

    @aeidon.deco.monkey_patch(gaupol.util, "flash_dialog")
    @aeidon.deco.silent(gaupol.Default)
    def test_close__translation_changed__discard(self):
        gaupol.util.flash_dialog = lambda *args: gtk.RESPONSE_NO
        doc = aeidon.documents.TRAN
        self.application.open_main(self.new_subrip_file())
        self.application.open_translation(self.new_microdvd_file())
        self.application.pages[-1].project.clear_texts((0,), doc)
        self.application.close(self.application.pages[-1], True)

    @aeidon.deco.monkey_patch(gaupol.util, "flash_dialog")
    @aeidon.deco.silent(gaupol.Default)
    def test_close__translation_changed__save(self):
        gaupol.util.flash_dialog = lambda *args: gtk.RESPONSE_YES
        doc = aeidon.documents.TRAN
        self.application.open_main(self.new_subrip_file())
        self.application.open_translation(self.new_microdvd_file())
        self.application.pages[-1].project.clear_texts((0,), doc)
        self.application.close(self.application.pages[-1], True)

    @aeidon.deco.monkey_patch(gaupol.util, "flash_dialog")
    @aeidon.deco.silent(gaupol.Default)
    def test_close__translation_removed(self):
        gaupol.util.flash_dialog = lambda *args: gtk.RESPONSE_CANCEL
        page = self.application.get_current_page()
        os.remove(page.project.tran_file.path)
        self.application.close(page, True)

    @aeidon.deco.silent(gaupol.Default)
    def test_close__unchanged__confirm(self):
        self.application.open_main(self.new_subrip_file())
        self.application.open_translation(self.new_microdvd_file())
        self.application.close(self.application.pages[-1], True)

    @aeidon.deco.silent(gaupol.Default)
    def test_close__unchanged__no_confirm(self):
        self.application.open_main(self.new_subrip_file())
        self.application.open_translation(self.new_microdvd_file())
        self.application.close(self.application.pages[-1], False)

    @aeidon.deco.monkey_patch(gaupol.util, "flash_dialog")
    @aeidon.deco.silent(gaupol.Default)
    def test_close_all__multiple(self):
        gaupol.util.flash_dialog = lambda *args: gtk.RESPONSE_NO
        self.application.open_main(self.new_subrip_file())
        self.application.open_main(self.new_subrip_file())
        for page in self.application.pages:
            page.project.remove_subtitles((0,))
        self.application.close_all()

    @aeidon.deco.monkey_patch(gaupol.util, "flash_dialog")
    @aeidon.deco.silent(gaupol.Default)
    def test_close_all__single(self):
        gaupol.util.flash_dialog = lambda *args: gtk.RESPONSE_NO
        self.application.open_main(self.new_subrip_file())
        self.application.pages[-1].project.remove_subtitles((0,))
        self.application.close_all()

    @aeidon.deco.monkey_patch(gaupol.util, "flash_dialog")
    def test_on_close_all_projects_activate(self):
        gaupol.util.flash_dialog = lambda *args: gtk.RESPONSE_NO
        self.application.get_action("close_all_projects").activate()
        self.application.open_main(self.new_subrip_file())
        self.application.open_main(self.new_subrip_file())
        for page in self.application.pages:
            page.project.remove_subtitles((0,))
        self.application.get_action("close_all_projects").activate()

    @aeidon.deco.monkey_patch(gaupol.util, "flash_dialog")
    def test_on_close_project_activate(self):
        gaupol.util.flash_dialog = lambda *args: gtk.RESPONSE_NO
        self.application.get_action("close_project").activate()
        self.application.open_main(self.new_subrip_file())
        self.application.pages[-1].project.remove_subtitles((0,))
        self.application.get_action("close_project").activate()

    @aeidon.deco.monkey_patch(gaupol.util, "flash_dialog")
    def test_on_page_close_request(self):
        gaupol.util.flash_dialog = lambda *args: gtk.RESPONSE_NO
        self.application.pages[-1].emit("close-request")
        self.application.open_main(self.new_subrip_file())
        self.application.pages[-1].project.remove_subtitles((0,))
        self.application.pages[-1].emit("close-request")

    @aeidon.deco.monkey_patch(gaupol.util, "flash_dialog")
    def test_on_quit_activate(self):
        gaupol.util.flash_dialog = lambda *args: gtk.RESPONSE_CANCEL
        for page in self.application.pages:
            page.project.remove_subtitles((0,))
        self.application.get_action("quit").activate()

    @aeidon.deco.monkey_patch(gaupol.util, "flash_dialog")
    def test_on_window_delete_event(self):
        gaupol.util.flash_dialog = lambda *args: gtk.RESPONSE_CANCEL
        for page in self.application.pages:
            page.project.remove_subtitles((0,))
        self.application.window.emit("delete-event", None)

    @aeidon.deco.monkey_patch(gaupol.util, "flash_dialog")
    def test_quit__default(self):
        gaupol.util.flash_dialog = lambda *args: gtk.RESPONSE_CANCEL
        for page in self.application.pages:
            page.project.remove_subtitles((0,))
        self.raises(gaupol.Default, self.application.quit)



# Copyright (C) 2005-2007,2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Internal text clipboard."""

__all__ = ("Clipboard",)


class Clipboard(object):

    """Internal text clipboard."""

    def __init__(self):
        """Initialize a Clipboard object."""
        # List of strings, each being the text of one subtitle. Nones in the
        # list express that those subtitles are skipped, i.e. the range of
        # subtitles is not unified, but contains gaps.
        self._texts = []

    def append(self, item):
        """Append `item` to texts."""
        self._texts.append(item)

    def clear(self):
        """Clear texts."""
        self._texts = []

    def get_string(self):
        """Return texts as one string."""
        strings = [x or "" for x in self._texts]
        return "\n\n".join(strings)

    def get_texts(self):
        """Return texts as a list of strings."""
        return self._texts[:]

    def set_texts(self, texts):
        """Set the list of texts."""
        self._texts = list(texts)

    def is_empty(self):
        """Return ``True`` if empty."""
        return not bool(self._texts)



# Copyright (C) 2006-2008,2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Functions to calculate line lengths and to show them in widgets."""

from __future__ import division

import aeidon
import gaupol
import gtk
import pango


class _Ruler(object):

    """Measurer of line lengths in various units.

    The "sans" font (whatever it happens to be) is used for size calculations
    in em units, as subtitles are commonly shown in a some sans serif font.
    This is an approximation, but should be far more accurate than using
    characters, at least in Latin and other variable width character scripts.
    """

    def __init__(self):
        """Initialize a :class:`_Ruler` object."""
        self._em_length = None
        self._layout = None
        self._length_unit = None
        self._update_em_length()
        self._update_length_unit()
        gaupol.conf.connect_notify("editor", "length_unit", self)

    def _on_conf_editor_notify_length_unit(self, *args):
        """Update the length function used."""
        self._update_length_unit()

    def _update_em_length(self):
        """Update the length of em based on font description size."""
        self._layout = gtk.Label().get_layout().copy()
        font_desc = self._layout.get_context().get_font_description()
        font_desc.merge(pango.FontDescription("sans"), True)
        self._layout.set_font_description(font_desc)
        self._em_length = font_desc.get_size()

    def _update_length_unit(self):
        """Update the length function used."""
        self._length_unit = gaupol.conf.editor.length_unit

    def get_char_length(self, text, strip=False, floor=False):
        """Return length of `text` measured in characters."""
        text = (aeidon.re_any_tag.sub("", text) if strip else text)
        return len(text.replace("\n", " "))

    def get_char_lengths(self, text, strip=False, floor=False):
        """Return line lengths of `text` measured in characters."""
        text = (aeidon.re_any_tag.sub("", text) if strip else text)
        return tuple(len(x) for x in text.split("\n"))

    def get_em_length(self, text, strip=False, floor=False):
        """Return length of `text` measured in ems."""
        text = (aeidon.re_any_tag.sub("", text) if strip else text)
        text = text.replace("\n", " ")
        self._layout.set_text(text)
        length = self._layout.get_size()[0] / self._em_length
        return (int(length) if floor else length)

    def get_em_lengths(self, text, strip=False, floor=False):
        """Return line lengths of `text` measured in ems."""
        text = (aeidon.re_any_tag.sub("", text) if strip else text)
        lengths = []
        for line in text.split("\n"):
            self._layout.set_text(line)
            length = self._layout.get_size()[0] / self._em_length
            lengths.append(int(length) if floor else length)
        return tuple(lengths)

    def get_lengths(self, text, strip=False, floor=False):
        """Return line lengths of `text` measured in default units."""
        if self._length_unit == gaupol.length_units.CHAR:
            return self.get_char_lengths(text, strip, floor)
        if self._length_unit == gaupol.length_units.EM:
            return self.get_em_lengths(text, strip, floor)
        raise ValueError("Invalid length unit: %s" % repr(self._length_unit))


_ruler = _Ruler()

def _on_text_view_expose_event(text_view, event):
    """Calculate and show line lengths in the margin."""
    text_buffer = text_view.get_buffer()
    bounds = text_buffer.get_bounds()
    text = text_buffer.get_text(*bounds)
    if not text: return
    lengths = get_lengths(text)
    layout = pango.Layout(text_view.get_pango_context())
    layout.set_markup("\n".join([str(x) for x in lengths]))
    layout.set_alignment(pango.ALIGN_RIGHT)
    width = layout.get_pixel_size()[0]
    text_view.set_border_window_size(gtk.TEXT_WINDOW_RIGHT, width + 4)
    y = -text_view.window_to_buffer_coords(gtk.TEXT_WINDOW_RIGHT, 2, 0)[1]
    window = text_view.get_window(gtk.TEXT_WINDOW_RIGHT)
    window.clear()
    text_view.style.paint_layout(window=window,
                                 state_type=gtk.STATE_NORMAL,
                                 use_text=True,
                                 area=None,
                                 widget=text_view,
                                 detail=None,
                                 x=2,
                                 y=y,
                                 layout=layout)

def connect_text_view(text_view):
    """Connect `text_view` to show line lengths in the margin."""
    context = text_view.get_pango_context()
    layout = pango.Layout(context)
    layout.set_text("8")
    width = layout.get_pixel_size()[0]
    text_view.set_border_window_size(gtk.TEXT_WINDOW_RIGHT, width + 4)
    handler_id = text_view.connect("expose-event", _on_text_view_expose_event)
    text_view.set_data("ruler_handler_id", handler_id)
    return handler_id

def disconnect_text_view(text_view):
    """Disconnect `text_view` from showing line lengths in the margin."""
    text_view.set_border_window_size(gtk.TEXT_WINDOW_RIGHT, 0)
    handler_id = text_view.get_data("ruler_handler_id")
    if handler_id is None: return
    text_view.set_data("ruler_handler_id", None)
    return text_view.disconnect(handler_id)

def get_length_function(unit):
    """Return a function that returns text length in `unit`."""
    if unit == gaupol.length_units.CHAR:
        return _ruler.get_char_length
    if unit == gaupol.length_units.EM:
        return _ruler.get_em_length
    raise ValueError("Invalid length unit: %s" % repr(unit))

def get_lengths(text):
    """Return a sequence of floored line lengths without tags."""
    return _ruler.get_lengths(text, True, True)



# Copyright (C) 2005-2007 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import gaupol
import gtk


class TestTimeEntry(gaupol.TestCase):

    def run__entry(self):
        self.entry.set_text("12:34:56.789")
        window = gtk.Window()
        window.connect("delete-event", gtk.main_quit)
        window.set_position(gtk.WIN_POS_CENTER)
        window.set_default_size(200, 50)
        window.add(self.entry)
        window.show_all()
        gtk.main()

    def setup_method(self, method):
        self.entry = gaupol.TimeEntry()

    def test_set_text(self):
        self.entry.set_text("12:34:56.789")
        self.entry.set_text("-12:34:56.789")
        self.entry.set_text("12:34:56,789")



# Copyright (C) 2005-2008,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""GTK+ user interface initialization."""

import aeidon
import atexit
import gaupol
import gettext
import locale
import optparse
import os
import re
import sys
_ = aeidon.i18n._


def _check_dependencies():
    """Check existance and versions of dependencies."""
    if sys.version_info[:3] < (2, 6, 0):
        print "Python 2.6 or greater is required to run Gaupol."
        raise SystemExit(1)
    try:
        import gtk
        if gtk.pygtk_version < (2, 16, 0):
            raise ImportError
    except ImportError:
        print "PyGTK 2.16 or greater is required to run Gaupol."
        raise SystemExit(1)

def _init_application(opts, args):
    """Initialize application and open files from `args`."""
    application = gaupol.Application()
    jump_row = None
    re_jump = re.compile(r"\+\d*")
    for arg in filter(re_jump.match, args):
        jump_row = (max(0, int(arg[1:]) - 1) if arg[1:] else -1)
        args.remove(arg)
    paths = map(os.path.abspath, args)
    if not opts.encoding in (None, "auto"):
        try: opts.encoding = aeidon.encodings.translate_code(opts.encoding)
        except ValueError:
            raise SystemExit("Unrecognized encoding: %s" % repr(opts.encoding))
    application.open_main(paths, opts.encoding)
    page = application.get_current_page()
    if (page is not None) and opts.translation_file:
        path = os.path.abspath(opts.translation_file)
        align_method = opts.align_method.upper()
        align_method = getattr(aeidon.align_methods, align_method)
        application.open_translation(path, opts.encoding, align_method)
    if (page is not None) and opts.video_file:
        path = os.path.abspath(opts.video_file)
        page.project.video_path = path
        application.update_gui()
    if (page is not None) and (jump_row is not None):
        page.view.set_focus(jump_row)
        page.view.scroll_to_row(jump_row)

def _init_configuration(path):
    """Read configuration values from file at `path`."""
    path = path or os.path.join(aeidon.CONFIG_HOME_DIR, "gaupol.conf")
    gaupol.conf.path = path
    gaupol.conf.read_from_file()
    atexit.register(gaupol.conf.write_to_file)

def _init_gettext():
    """Initialize translation settings."""
    locale.setlocale(locale.LC_ALL, "")
    try:
        # Not available on all platforms.
        locale.bindtextdomain("gaupol", aeidon.LOCALE_DIR)
        locale.textdomain("gaupol")
    except AttributeError: pass
    gettext.bindtextdomain("gaupol", aeidon.LOCALE_DIR)
    gettext.textdomain("gaupol")

def _on_parser_list_encodings(*args):
    """List all available character encodings and exit."""
    encodings = [x[0] for x in aeidon.encodings.get_valid()]
    if aeidon.util.chardet_available():
        encodings.insert(0, "auto")
    print "\n".join(encodings)
    raise SystemExit(0)

def _on_parser_version(*args):
    """Show the version number and exit."""
    print "gaupol %s" % gaupol.__version__
    raise SystemExit(0)

def _parse_args(args):
    """Parse and return options and arguments from `args`."""
    parser = optparse.OptionParser(
        formatter=optparse.IndentedHelpFormatter(2, 42),
        usage=_("gaupol [OPTION...] [FILE...] [+[NUM]]"))

    parser.add_option(
        "--version",
        action="callback",
        callback=_on_parser_version,
        help=_("show version number and exit"),)

    parser.add_option(
        "-c", "--config-file",
        action="store",
        type="string",
        metavar=_("FILE"),
        dest="config_file",
        default=None,
        help=_("set the configuration file used"),)

    parser.add_option(
        "-e", "--encoding",
        action="store",
        type="string",
        metavar=_("ENCODING"),
        dest="encoding",
        default=None,
        help=_("set the character encoding used to open files"),)

    parser.add_option(
        "--list-encodings",
        action="callback",
        callback=_on_parser_list_encodings,
        help=_("list all available character encodings"),)

    parser.add_option(
        "-t", "--translation-file",
        action="store",
        type="string",
        metavar=_("FILE"),
        dest="translation_file",
        default=None,
        help=_("open translation file"),)

    parser.add_option(
        "-a", "--align-method",
        action="store",
        type="string",
        metavar=_("METHOD"),
        dest="align_method",
        default="position",
        help=_("method used to align translation subtitles: "
               "'number' or 'position'"),)

    parser.add_option(
        "-v", "--video-file",
        action="store",
        type="string",
        metavar=_("FILE"),
        dest="video_file",
        default=None,
        help=_("select video file"),)

    return parser.parse_args(args)

def main(args):
    """Parse arguments from `args` and start application."""
    _check_dependencies()
    aeidon.paths.xdg_copy_if_applicable()
    opts, args = _parse_args(args)
    sys.excepthook = gaupol.util.show_exception
    _init_gettext()
    _init_configuration(opts.config_file)
    _init_application(opts, args)
    import gtk
    gtk.main()



# Copyright (C) 2005-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import aeidon

MAIN = aeidon.documents.MAIN
TRAN = aeidon.documents.TRAN


class TestRegisterAgent(aeidon.TestCase):

    def setup_method(self, method):
        self.project = self.new_project()
        self.delegate = self.project.undo.im_self

    def test__on_notify_undo_limit(self):
        self.project.clear_texts((0,), MAIN)
        self.project.clear_texts((1,), MAIN)
        assert len(self.project.undoables) == 2
        self.project.undo_limit = 1
        assert len(self.project.undoables) == 1

    def test_can_redo(self):
        assert not self.project.can_redo()
        self.project.clear_texts((0,), MAIN)
        assert not self.project.can_redo()
        self.project.undo()
        assert self.project.can_redo()

    def test_can_undo(self):
        assert not self.project.can_undo()
        self.project.clear_texts((0,), MAIN)
        assert self.project.can_undo()
        self.project.undo()
        assert not self.project.can_undo()

    def test_cut_reversion_stacks(self):
        self.project.clear_texts((0,), MAIN)
        self.project.clear_texts((1,), MAIN)
        self.project.cut_reversion_stacks()
        self.project.undo_limit = 1
        self.project.cut_reversion_stacks()

    def test_emit_action_signal(self):
        self.project.clear_texts((0,), MAIN)
        self.project.emit_action_signal(aeidon.registers.DO)

    def test_group_actions(self):
        self.project.clear_texts((0,), MAIN)
        self.project.clear_texts((1,), MAIN)
        self.project.clear_texts((2,), MAIN)
        self.project.group_actions(aeidon.registers.DO, 2, "")
        assert len(self.project.undoables) == 2
        assert self.project.undoables[0].description == ""
        assert len(self.project.undoables[0].actions) == 2

    def test_redo(self):
        self.project.clear_texts((0,), MAIN)
        self.project.clear_texts((1,), MAIN)
        self.project.clear_texts((2,), MAIN)
        self.project.undo(3)
        self.project.redo(1)
        self.project.redo(2)

    def test_redo__multiple(self):
        for i in range(6):
            self.project.clear_texts([i], MAIN)
        self.project.group_actions(aeidon.registers.DO, 3, "")
        self.project.remove_subtitles((3, 4))
        self.project.insert_subtitles((3,))
        self.project.undo(6)
        self.project.redo(6)

    def test_register_action(self):
        self.project.clear_texts((0,), TRAN)
        self.project.clear_texts((1,), MAIN)
        self.project.clear_texts((2,), MAIN)
        self.project.undo(3)
        self.project.redo(2)
        assert len(self.project.undoables) == 2
        assert len(self.project.redoables) == 1

    def test_register_action__activate_translation(self):
        path = self.new_subrip_file()
        self.project.open_main(path, "ascii")
        assert self.project.tran_changed is None
        self.project.clear_texts((1,), TRAN)
        assert self.project.tran_changed == 1

    def test_set_action_description(self):
        self.project.clear_texts((0,), MAIN)
        self.project.set_action_description(aeidon.registers.DO, "")
        assert self.project.undoables[0].description == ""

    def test_undo(self):
        self.project.clear_texts((0,), MAIN)
        self.project.clear_texts((1,), MAIN)
        self.project.clear_texts((2,), MAIN)
        self.project.undo(1)
        self.project.undo(2)
        self.project.redo(3)

    def test_undo__multiple(self):
        for i in range(6):
            self.project.clear_texts([i], MAIN)
        self.project.group_actions(aeidon.registers.DO, 3, "")
        self.project.remove_subtitles((3, 4))
        self.project.insert_subtitles((3,))
        self.project.undo(6)
        self.project.redo(6)



# Copyright (C) 2005-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import aeidon


class TestModule(aeidon.TestCase):

    def test_add(self):
        aeidon.files.add(aeidon.files.SubRip)

    def test_new(self):
        format = aeidon.formats.SUBRIP
        file = aeidon.files.new(format, "", "ascii")
        assert isinstance(file, aeidon.files.SubRip)

    def test_new__value_error(self):
        self.raises(ValueError, aeidon.files.new, None, "", "ascii")



# Copyright (C) 2005-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import aeidon


class TestSubStationAlpha(aeidon.TestCase):

    format = aeidon.formats.SSA

    def setup_method(self, method):
        self.file = aeidon.files.new(self.format,
                                     self.new_temp_file(self.format),
                                     "ascii")

    def test_copy_from(self):
        self.file.header = "test"
        self.file.event_fields = ("Marked",)
        path = self.new_temp_file(self.format)
        new_file = aeidon.files.new(self.format, path, "ascii")
        new_file.copy_from(self.file)
        assert new_file.header == "test"
        assert new_file.event_fields == ("Marked",)

    def test_read(self):
        assert self.file.read()
        assert self.file.header

    def test_write(self):
        self.file.write(self.file.read(), aeidon.documents.MAIN)
        text = open(self.file.path, "r").read().strip()
        reference = self.get_sample_text(self.format)
        assert text == reference
